#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core import debug
from fenrirscreenreader.core.i18n import _


class HelpManager:
    def __init__(self):
        self.helpDict = {}
        self.tutorialListIndex = None

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def toggle_tutorial_mode(self):
        self.set_tutorial_mode(not self.env["general"]["tutorialMode"])

    def set_tutorial_mode(self, newTutorialMode):
        if self.env["runtime"]["VmenuManager"].get_active():
            return
        self.env["general"]["tutorialMode"] = newTutorialMode
        if newTutorialMode:
            self.create_help_dict()
            self.env["bindings"][
                str([1, ["KEY_ESC"]])
            ] = "TOGGLE_TUTORIAL_MODE"
            self.env["bindings"][str([1, ["KEY_UP"]])] = "PREV_HELP"
            self.env["bindings"][str([1, ["KEY_DOWN"]])] = "NEXT_HELP"
            self.env["bindings"][str([1, ["KEY_SPACE"]])] = "CURR_HELP"
        else:
            try:
                self.env["bindings"] = self.env["runtime"][
                    "SettingsManager"
                ].get_binding_backup()
            except Exception as e:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "HelpManager set_tutorial_mode: Error restoring binding backup: "
                    + str(e),
                    debug.DebugLevel.ERROR,
                )

    def is_tutorial_mode(self):
        return self.env["general"]["tutorialMode"]

    def get_formatted_shortcut_for_command(self, command):
        shortcut = []
        raw_shortcut = []
        try:
            raw_shortcut = list(self.env["bindings"].keys())[
                list(self.env["bindings"].values()).index(command)
            ]
            raw_shortcut = self.env["rawBindings"][raw_shortcut]
            # prefer numbers for multitap
            if raw_shortcut[0] in range(2, 9):
                formatted_key = str(raw_shortcut[0]) + " times "
                shortcut.append(formatted_key)
            # prefer metha keys
            for k in [
                "KEY_FENRIR",
                "KEY_SCRIPT",
                "KEY_CTRL",
                "KEY_SHIFT",
                "KEY_ALT",
                "KEY_META",
            ]:
                if k in raw_shortcut[1]:
                    formatted_key = k
                    formatted_key = formatted_key.lower()
                    formatted_key = formatted_key.replace("key_kp", " keypad ")
                    formatted_key = formatted_key.replace("key_", " ")
                    shortcut.append(formatted_key)
                    raw_shortcut[1].remove(k)
            # handle other keys
            for k in raw_shortcut[1]:
                formatted_key = k
                formatted_key = formatted_key.lower()
                formatted_key = formatted_key.replace("key_kp", " keypad ")
                formatted_key = formatted_key.replace("key_", " ")
                shortcut.append(formatted_key)
        except Exception as e:
            return ""
        shortcut = str(shortcut)
        shortcut = shortcut.replace("[", "")
        shortcut = shortcut.replace("]", "")
        shortcut = shortcut.replace("'", "")
        return shortcut

    def get_command_help_text(self, command, section="commands"):
        command_name = command.lower()
        command_name = command_name.split("__-__")[0]
        command_name = command_name.replace("_", " ")
        command_name = command_name.replace("_", " ")
        if command == "TOGGLE_TUTORIAL_MODE":
            command_description = _("toggles the tutorial mode")
        else:
            command_description = self.env["runtime"][
                "CommandManager"
            ].get_command_description(command, section="commands")
        if command_description == "":
            command_description = "no Description available"
        command_shortcut = self.get_formatted_shortcut_for_command(command)
        if command_shortcut == "":
            command_shortcut = "unbound"
        helptext = (
            command_name
            + ", Shortcut "
            + command_shortcut
            + ", Description "
            + command_description
        )
        return helptext

    def create_help_dict(self, section="commands"):
        self.helpDict = {}
        for command in sorted(self.env["commands"][section].keys()):
            self.helpDict[len(self.helpDict)] = self.get_command_help_text(
                command, section
            )
        if len(self.helpDict) > 0:
            self.tutorialListIndex = 0
        else:
            self.tutorialListIndex = None

    def get_help_for_current_index(self):
        if self.tutorialListIndex is None:
            return ""
        return self.helpDict[self.tutorialListIndex]

    def next_index(self):
        if self.tutorialListIndex is None:
            return
        self.tutorialListIndex += 1
        if self.tutorialListIndex >= len(self.helpDict):
            self.tutorialListIndex = 0

    def prev_index(self):
        if self.tutorialListIndex is None:
            return
        self.tutorialListIndex -= 1
        if self.tutorialListIndex < 0:
            self.tutorialListIndex = len(self.helpDict) - 1



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core.i18n import _
from fenrirscreenreader.utils import line_utils
from fenrirscreenreader.utils import word_utils


class command:
    def __init__(self):
        self.lastIdent = -1

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return ""

    def run(self):
        if not self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "focus", "cursor"
        ):
            return
        if self.env["runtime"]["ScreenManager"].is_screen_change():
            self.lastIdent = 0
            return
        # Don't announce cursor movements when auto-read is handling incoming text
        # This prevents interrupting ongoing auto-read announcements
        if self.env["runtime"]["ScreenManager"].is_delta():
            return

        # is a vertical change?
        if not self.env["runtime"]["CursorManager"].is_cursor_vertical_move():
            return

        x, y, curr_line = line_utils.get_current_line(
            self.env["screen"]["new_cursor"]["x"],
            self.env["screen"]["new_cursor"]["y"],
            self.env["screen"]["new_content_text"],
        )
        # Don't interrupt ongoing auto-read announcements with cursor movement
        do_interrupt = True
        if self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "speech", "auto_read_incoming"
        ):
            do_interrupt = False

        if curr_line.isspace():
            self.env["runtime"]["OutputManager"].present_text(
                _("blank"), sound_icon="EmptyLine", interrupt=do_interrupt, flush=False
            )
        else:
            # ident
            curr_ident = len(curr_line) - len(curr_line.lstrip())
            if self.lastIdent == -1:
                self.lastIdent = curr_ident
            if self.env["runtime"]["SettingsManager"].get_setting_as_bool(
                "general", "auto_present_indent"
            ):
                if self.env["runtime"]["SettingsManager"].get_setting_as_int(
                    "general", "auto_present_indent_mode"
                ) in [0, 1]:
                    if self.lastIdent != curr_ident:
                        self.env["runtime"]["OutputManager"].play_frequence(
                            curr_ident * 50, 0.1, interrupt=do_interrupt
                        )
                if self.env["runtime"]["SettingsManager"].get_setting_as_int(
                    "general", "auto_present_indent_mode"
                ) in [0, 2]:
                    if self.lastIdent != curr_ident:
                        self.env["runtime"]["OutputManager"].present_text(
                            _("indented ") + str(curr_ident) + " ",
                            interrupt=do_interrupt,
                            flush=False,
                        )
                        do_interrupt = False
            # barrier
            say_line = curr_line
            if self.env["runtime"]["SettingsManager"].get_setting_as_bool(
                "barrier", "enabled"
            ):
                is_barrier, barrierLine = self.env["runtime"][
                    "BarrierManager"
                ].handle_line_barrier(
                    self.env["screen"]["new_content_text"].split("\n"),
                    self.env["screen"]["new_cursor"]["x"],
                    self.env["screen"]["new_cursor"]["y"],
                )
                if is_barrier:
                    say_line = barrierLine
            # output
            self.env["runtime"]["OutputManager"].present_text(
                say_line, interrupt=do_interrupt, flush=False
            )
            self.lastIdent = curr_ident

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import string

from fenrirscreenreader.core.i18n import _
from fenrirscreenreader.utils import word_utils


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return "No Description found"

    def run(self):
        # is navigation?
        if (
            not abs(
                self.env["screen"]["old_cursor"]["x"]
                - self.env["screen"]["new_cursor"]["x"]
            )
            > 1
        ):
            return

        # just when cursor move worddetection is needed
        if not self.env["runtime"][
            "CursorManager"
        ].is_cursor_horizontal_move():
            return
        # for now no new line
        if self.env["runtime"]["CursorManager"].is_cursor_vertical_move():
            return
        # currently writing
        if self.env["runtime"]["ScreenManager"].is_delta():
            return

        # get the word
        new_content = self.env["screen"]["new_content_text"].split("\n")[
            self.env["screen"]["new_cursor"]["y"]
        ]
        x, y, curr_word, end_of_screen, line_break = (
            word_utils.get_current_word(
                self.env["screen"]["new_cursor"]["x"], 0, new_content
            )
        )

        # is there a word?
        if curr_word == "":
            return

        # at the start of a word
        if (x + len(curr_word) != self.env["screen"]["new_cursor"]["x"]) and (
            self.env["screen"]["new_cursor"]["x"] != x
        ):
            return

        # Don't interrupt ongoing auto-read announcements
        do_interrupt = True
        if self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "speech", "auto_read_incoming"
        ):
            do_interrupt = False

        self.env["runtime"]["OutputManager"].present_text(
            curr_word, interrupt=do_interrupt, flush=False
        )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.
# attrib:
# http://rampex.ihep.su/Linux/linux_howto/html/tutorials/mini/Colour-ls-6.html
# 0 = black, 1 = blue, 2 = green, 3 = cyan, 4 = red, 5 = purple, 6 = brown/yellow, 7 = white.
# https://github.com/jwilk/vcsapeek/blob/master/linuxvt.py
# blink = 5 if attr & 1 else 0
# bold = 1 if attr & 16 else 0

import fcntl
import glob
import os
import select
import subprocess
import termios
import time
from array import array
from fcntl import ioctl
from struct import pack
from struct import unpack
from struct import unpack_from

import dbus

from fenrirscreenreader.core import debug
from fenrirscreenreader.core.eventData import FenrirEventType
from fenrirscreenreader.core.screenDriver import ScreenDriver as screenDriver
from fenrirscreenreader.utils import screen_utils


class driver(screenDriver):
    """Linux VCSA (Virtual Console Screen Access) driver for Fenrir screen reader.
    
    This driver provides access to Linux virtual consoles (TTYs) through the VCSA
    interface, allowing real-time monitoring of screen content and cursor position.
    It supports both text content extraction and color/attribute detection.
    
    The driver monitors multiple virtual consoles simultaneously and can detect:
    - Screen content changes (text updates)
    - Cursor movement
    - TTY switching
    - Text attributes (colors, bold, etc.)
    - Session information via D-Bus/logind
    
    Attributes:
        ListSessions: D-Bus method for listing login sessions
        sysBus: D-Bus system bus connection
        charmap: Character mapping for text decoding
        bgColorValues: Background color value mappings
        fgColorValues: Foreground color value mappings
        hichar: High character mask for Unicode support
    """
    def __init__(self):
        screenDriver.__init__(self)
        self.ListSessions = None
        self.sysBus = None
        self.charmap = {}
        self.bgColorValues = {
            0: "black",
            1: "blue",
            2: "green",
            3: "cyan",
            4: "red",
            5: "magenta",
            6: "brown/yellow",
            7: "white",
        }
        self.fgColorValues = {
            0: "black",
            1: "blue",
            2: "green",
            3: "cyan",
            4: "red",
            5: "magenta",
            6: "brown/yellow",
            7: "light gray",
            8: "dark gray",
            9: "light blue",
            10: "light green",
            11: "light cyan",
            12: "light red",
            13: "light magenta",
            14: "light yellow",
            15: "white",
        }
        self.hichar = None
        try:
            # set workaround for paste clipboard -> inject_text_to_screen
            subprocess.run(
                ["sysctl", "dev.tty.legacy_tiocsti=1"],
                check=False,
                capture_output=True,
                timeout=5,
            )
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "vcsaDriver shutdown: Error running fgconsole: " + str(e),
                debug.DebugLevel.ERROR,
            )

    def initialize(self, environment):
        """Initialize the VCSA driver with the given environment.
        
        Sets up default attributes, starts the screen monitoring watchdog process,
        and prepares the driver for screen content monitoring.
        
        Args:
            environment: The Fenrir environment dictionary containing runtime managers
                       and configuration settings.
        """
        self.env = environment
        self.env["runtime"]["AttributeManager"].append_default_attributes(
            [
                self.fgColorValues[7],  # fg
                self.bgColorValues[0],  # bg
                False,  # bold
                False,  # italics
                False,  # underscore
                False,  # strikethrough
                False,  # reverse
                False,  # blink
                "default",  # fontsize
                "default",  # fontfamily
            ]
        )  # end attribute )
        # Use threading instead of multiprocessing to avoid pickle issues
        # with self.env (which contains unpicklable file handles)
        self.env["runtime"]["ProcessManager"].add_custom_event_thread(
            self.update_watchdog, multiprocess=False
        )

    def get_curr_screen(self):
        """Get the currently active TTY number.
        
        Reads from /sys/devices/virtual/tty/tty0/active to determine which
        virtual console is currently active and updates the environment.
        
        Updates:
            env['screen']['oldTTY']: Previous TTY number
            env['screen']['newTTY']: Current TTY number
        """
        self.env["screen"]["oldTTY"] = self.env["screen"]["newTTY"]
        try:
            with open(
                "/sys/devices/virtual/tty/tty0/active", "r"
            ) as currScreenFile:
                self.env["screen"]["newTTY"] = str(currScreenFile.read()[3:-1])
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                str(e), debug.DebugLevel.ERROR
            )

    def inject_text_to_screen(self, text, screen=None):
        """Inject text into the specified screen as if typed by user.
        
        Uses the TIOCSTI ioctl to simulate keystrokes on the target TTY.
        This is primarily used for clipboard paste functionality.
        
        Args:
            text (str): Text to inject into the screen
            screen (str, optional): Target screen device (e.g., '/dev/tty1').
                                  If None, uses current TTY.
        
        Note:
            Requires appropriate permissions and may need legacy_tiocsti=1
            kernel parameter on newer systems.
        """
        use_screen = "/dev/tty" + self.env["screen"]["newTTY"]
        if screen is not None:
            use_screen = screen
        with open(use_screen, "w") as fd:
            text_bytes = text.encode('utf-8')
            for byte in text_bytes:
                fcntl.ioctl(fd, termios.TIOCSTI, bytes([byte]))

    def get_session_information(self):
        """Retrieve session information via D-Bus logind interface.
        
        Connects to systemd-logind to gather information about active sessions,
        including session types and TTY assignments. This helps identify which
        screens should be automatically ignored (e.g., X11 sessions).
        
        Updates:
            env['screen']['autoIgnoreScreens']: List of screens to ignore
            env['general']['curr_user']: Current user for active session
            env['general']['prev_user']: Previous user
        
        Note:
            Gracefully handles cases where logind is not available.
        """
        self.env["screen"]["autoIgnoreScreens"] = []
        try:
            if not self.sysBus:
                self.sysBus = dbus.SystemBus()
                obj = self.sysBus.get_object(
                    "org.freedesktop.login1", "/org/freedesktop/login1"
                )
                inf = dbus.Interface(obj, "org.freedesktop.login1.Manager")
                self.ListSessions = inf.get_dbus_method("ListSessions")
            sessions = self.ListSessions()

            for session in sessions:
                obj = self.sysBus.get_object(
                    "org.freedesktop.login1", session[4]
                )
                inf = dbus.Interface(obj, "org.freedesktop.DBus.Properties")
                session_type = inf.Get(
                    "org.freedesktop.login1.Session", "Type"
                )
                screen = str(inf.Get("org.freedesktop.login1.Session", "VTNr"))
                if screen == "":
                    screen = str(
                        inf.Get("org.freedesktop.login1.Session", "TTY")
                    )
                    screen = screen[screen.upper().find("TTY") + 3 :]
                if screen == "":
                    self.env["runtime"]["DebugManager"].write_debug_out(
                        "No TTY found for session:" + session[4],
                        debug.DebugLevel.ERROR,
                    )
                    return
                if session_type.upper() != "TTY":
                    self.env["screen"]["autoIgnoreScreens"] += [screen]
                if screen == self.env["screen"]["newTTY"]:
                    if self.env["general"]["curr_user"] != session[2]:
                        self.env["general"]["prev_user"] = self.env["general"][
                            "curr_user"
                        ]
                        self.env["general"]["curr_user"] = session[2]
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "get_session_information: Maybe no LoginD:" + str(e),
                debug.DebugLevel.ERROR,
            )
        # self.env['runtime']['DebugManager'].write_debug_out('get_session_information:'  + str(self.env['screen']['autoIgnoreScreens']) + ' ' + str(self.env['general'])  ,debug.DebugLevel.INFO)

    def read_file(self, file):
        """Read content from a file handle with error recovery.
        
        Attempts to read the entire file content, falling back to
        line-by-line reading if the initial read fails. This is used
        for reading VCSA/VCSU device files.
        
        Args:
            file: Open file handle to read from
            
        Returns:
            bytes: File content as bytes
        """
        d = b""
        file.seek(0)
        try:
            d = file.read()
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "vcsaDriver get_screen_text: Error reading file: " + str(e),
                debug.DebugLevel.ERROR,
            )
            file.seek(0)
            while True:
                # Read from file
                try:
                    d += file.readline(1)
                    if not d:
                        break
                except Exception as e:
                    break
        return d

    def update_watchdog(self, active, event_queue):
        """Main watchdog loop for monitoring screen changes.
        
        This is the core monitoring function that runs in a separate process.
        It uses epoll to efficiently monitor multiple VCSA devices and the
        active TTY file for changes. When changes are detected, it generates
        appropriate events for the main Fenrir process.
        
        The watchdog monitors:
        - Screen content changes (text updates)
        - TTY switches (screen changes)
        - Cursor position changes
        
        Args:
            active: Shared boolean value controlling the watchdog loop
            event_queue: Queue for sending events to the main process
            
        Events Generated:
            - FenrirEventType.screen_changed: When switching TTYs
            - FenrirEventType.screen_update: When screen content changes
            
        Note:
            This method runs in a multiprocess context and includes comprehensive
            cleanup of file handles in the finally block.
        """
        vcsa = {}
        vcsu = {}
        tty = None
        watchdog = None
        try:
            use_vcsu = os.access("/dev/vcsu", os.R_OK)
            vcsa_devices = glob.glob("/dev/vcsa*")
            vcsu_devices = None
            last_screen_content = b""

            # Open TTY file with proper cleanup
            tty = open("/sys/devices/virtual/tty/tty0/active", "r")
            curr_screen = str(tty.read()[3:-1])
            old_screen = curr_screen

            # Open VCSA devices with proper cleanup tracking
            for vcsaDev in vcsa_devices:
                index = str(vcsaDev[9:])
                vcsa[index] = open(vcsaDev, "rb")
                if index == curr_screen:
                    last_screen_content = self.read_file(vcsa[index])

            # Open VCSU devices if available
            if use_vcsu:
                vcsu_devices = glob.glob("/dev/vcsu*")
                for vcsuDev in vcsu_devices:
                    index = str(vcsuDev[9:])
                    vcsu[index] = open(vcsuDev, "rb")

            self.update_char_map(curr_screen)
            watchdog = select.epoll()
            watchdog.register(
                vcsa[curr_screen], select.POLLPRI | select.POLLERR
            )
            watchdog.register(tty, select.POLLPRI | select.POLLERR)
            while active.value:
                changes = watchdog.poll(1)
                for change in changes:
                    fileno = change[0]
                    event = change[1]
                    if fileno == tty.fileno():
                        self.env["runtime"]["DebugManager"].write_debug_out(
                            "ScreenChange", debug.DebugLevel.INFO
                        )
                        tty.seek(0)
                        curr_screen = str(tty.read()[3:-1])
                        if curr_screen != old_screen:
                            try:
                                watchdog.unregister(vcsa[old_screen])
                            except Exception as e:
                                self.env["runtime"][
                                    "DebugManager"
                                ].write_debug_out(
                                    "vcsaDriver update_watchdog: Error unregistering watchdog: "
                                    + str(e),
                                    debug.DebugLevel.ERROR,
                                )
                            try:
                                watchdog.register(
                                    vcsa[curr_screen],
                                    select.POLLPRI | select.POLLERR,
                                )
                            except Exception as e:
                                self.env["runtime"][
                                    "DebugManager"
                                ].write_debug_out(
                                    "vcsaDriver update_watchdog: Error registering watchdog: "
                                    + str(e),
                                    debug.DebugLevel.ERROR,
                                )
                            self.update_char_map(curr_screen)
                            old_screen = curr_screen
                            try:
                                vcsa[curr_screen].seek(0)
                                last_screen_content = self.read_file(
                                    vcsa[curr_screen]
                                )
                            except Exception as e:
                                self.env["runtime"][
                                    "DebugManager"
                                ].write_debug_out(
                                    "vcsaDriver update_watchdog: Error reading screen content: "
                                    + str(e),
                                    debug.DebugLevel.ERROR,
                                )
                            vcsu_content = None
                            if use_vcsu:
                                vcsu[curr_screen].seek(0)
                                vcsu_content = self.read_file(
                                    vcsu[curr_screen]
                                )
                            event_queue.put(
                                {
                                    "Type": FenrirEventType.screen_changed,
                                    "data": self.create_screen_event_data(
                                        curr_screen,
                                        last_screen_content,
                                        vcsu_content,
                                    ),
                                }
                            )
                    else:
                        self.env["runtime"]["DebugManager"].write_debug_out(
                            "screen_update", debug.DebugLevel.INFO
                        )
                        vcsa[curr_screen].seek(0)
                        time.sleep(0.01)
                        dirty_content = self.read_file(vcsa[curr_screen])
                        screen_content = dirty_content
                        vcsu_content = None
                        timeout = time.time()
                        # error case
                        if screen_content == b"":
                            continue
                        if last_screen_content == b"":
                            last_screen_content = screen_content
                        if (
                            abs(
                                int(screen_content[2])
                                - int(last_screen_content[2])
                            )
                            in [1, 2]
                        ) and (
                            int(screen_content[3])
                            == int(last_screen_content[3])
                        ):
                            # Skip X Movement
                            pass
                        elif (
                            abs(
                                int(screen_content[3])
                                - int(last_screen_content[3])
                            )
                            in [1]
                        ) and (
                            int(screen_content[2])
                            == int(last_screen_content[2])
                        ):
                            # Skip Y Movement
                            pass
                        else:
                            # anything else? wait for completion
                            while True:
                                screen_content = dirty_content
                                time.sleep(0.02)
                                # r,_,_ = select.select([vcsa[curr_screen]], [], [], 0.07)
                                # if not vcsa[curr_screen] in r:
                                #    break
                                vcsa[curr_screen].seek(0)
                                dirty_content = self.read_file(
                                    vcsa[curr_screen]
                                )
                                if screen_content == dirty_content:
                                    break
                                if time.time() - timeout >= 0.1:
                                    screen_content = dirty_content
                                    break
                        if use_vcsu:
                            vcsu[curr_screen].seek(0)
                            vcsu_content = self.read_file(vcsu[curr_screen])
                        last_screen_content = screen_content
                        event_queue.put(
                            {
                                "Type": FenrirEventType.screen_update,
                                "data": self.create_screen_event_data(
                                    curr_screen, screen_content, vcsu_content
                                ),
                            }
                        )
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "VCSA:update_watchdog:" + str(e), debug.DebugLevel.ERROR
            )
            time.sleep(0.2)
        finally:
            # Clean up all file handles
            try:
                if watchdog:
                    watchdog.close()
            except Exception as e:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "vcsaDriver update_watchdog: Error closing watchdog: "
                    + str(e),
                    debug.DebugLevel.ERROR,
                )
            try:
                if tty:
                    tty.close()
            except Exception as e:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "vcsaDriver shutdown: Error closing TTY: " + str(e),
                    debug.DebugLevel.ERROR,
                )
            for handle in vcsa.values():
                try:
                    handle.close()
                except Exception as e:
                    self.env["runtime"]["DebugManager"].write_debug_out(
                        "vcsaDriver shutdown: Error closing VCSA handle: "
                        + str(e),
                        debug.DebugLevel.ERROR,
                    )
            for handle in vcsu.values():
                try:
                    handle.close()
                except Exception as e:
                    self.env["runtime"]["DebugManager"].write_debug_out(
                        "vcsaDriver shutdown: Error closing VCSU handle: "
                        + str(e),
                        debug.DebugLevel.ERROR,
                    )

    def create_screen_event_data(self, screen, vcsaContent, vcsu_content=None):
        """Create standardized screen event data from VCSA content.
        
        Processes raw VCSA bytes into a structured event data dictionary
        containing screen dimensions, cursor position, text content, and
        color attributes.
        
        Args:
            screen (str): TTY number (e.g., '1' for tty1)
            vcsaContent (bytes): Raw VCSA device content
            vcsu_content (bytes, optional): VCSU content for Unicode support
            
        Returns:
            dict: Event data with keys:
                - bytes: Raw VCSA content
                - lines: Screen height
                - columns: Screen width  
                - textCursor: Cursor position {x, y}
                - screen: TTY number
                - screenUpdateTime: Timestamp
                - text: Decoded text content
                - attributes: Color/formatting attributes
        """
        event_data = {
            "bytes": vcsaContent,
            "lines": int(vcsaContent[0]),
            "columns": int(vcsaContent[1]),
            "textCursor": {"x": int(vcsaContent[2]), "y": int(vcsaContent[3])},
            "screen": screen,
            "screenUpdateTime": time.time(),
            "text": "",
            "attributes": [],
        }
        try:
            event_data["text"], event_data["attributes"] = (
                self.auto_decode_vcsa(
                    vcsaContent[4:], event_data["lines"], event_data["columns"]
                )
            )
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "vcsaDriver create_screen_event_data: Error decoding VCSA content: "
                + str(e),
                debug.DebugLevel.ERROR,
            )
        # VCSU seems to give b'   ' instead of b'\x00\x00\x00' (tsp),
        # deactivated until its fixed
        if vcsu_content is not None:
            try:
                vcsu_content_as_text = vcsu_content.decode("UTF-32")
                event_data["text"] = screen_utils.insert_newlines(
                    vcsu_content_as_text, event_data["columns"]
                )
            except Exception as e:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "vcsaDriver create_screen_event_data: Error decoding VCSU content: "
                    + str(e),
                    debug.DebugLevel.ERROR,
                )
        return event_data.copy()

    def update_char_map(self, screen):
        """Update character mapping for the specified screen.
        
        Reads the Unicode font mapping from the TTY to properly decode
        character data from VCSA. This handles special characters and
        Unicode properly.
        
        Args:
            screen (str): TTY number to update mapping for
            
        Updates:
            self.charmap: Dictionary mapping byte values to Unicode characters
            self.hichar: High character mask for extended characters
        """
        self.charmap = {}
        try:
            with open("/dev/tty" + screen, "rb") as tty:
                GIO_UNIMAP = 0x4B66
                VT_GETHIFONTMASK = 0x560D
                himask = array("H", (0,))
                ioctl(tty, VT_GETHIFONTMASK, himask)
                (self.hichar,) = unpack_from("@H", himask)
                sz = 512
                line = ""
                while True:
                    try:
                        unipairs = array("H", [0] * (2 * sz))
                        unimapdesc = array(
                            "B", pack("@HP", sz, unipairs.buffer_info()[0])
                        )
                        ioctl(tty.fileno(), GIO_UNIMAP, unimapdesc)
                        break
                    except Exception as e:
                        self.env["runtime"]["DebugManager"].write_debug_out(
                            "VCSA:update_char_map:scaling up sz="
                            + str(sz)
                            + " "
                            + str(e),
                            debug.DebugLevel.WARNING,
                        )
                        sz *= 2
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "VCSA:update_char_map:" + str(e), debug.DebugLevel.ERROR
            )
            return
        (ncodes,) = unpack_from("@H", unimapdesc)
        utable = unpack_from("@%dH" % (2 * ncodes), unipairs)
        for u, b in zip(utable[::2], utable[1::2]):
            if self.charmap.get(b) is None:
                self.charmap[b] = chr(u)

    def auto_decode_vcsa(self, allData, rows, cols):
        """Decode raw VCSA data into text and attributes.
        
        Processes the character and attribute data from VCSA devices,
        extracting both the text content and formatting information
        (colors, bold, blink, etc.).
        
        Args:
            allData (bytes): Raw character and attribute data from VCSA
            rows (int): Number of screen rows
            cols (int): Number of screen columns
            
        Returns:
            tuple: (text_content, attributes)
                - text_content (str): Decoded text with newlines
                - attributes (list): List of attribute arrays for each character
                
        Note:
            Each character in VCSA is stored as 2 bytes: character + attribute.
            Attributes encode foreground/background colors, bold, blink, etc.
        """
        all_text = ""
        all_attrib = []
        i = 0
        for y in range(rows):
            line_text = ""
            line_attrib = []
            blink = 0
            bold = 0
            ink = 7
            paper = 0
            for x in range(cols):
                data = allData[i : i + 2]
                i += 2
                if data == b" \x07":
                    # attr = 7
                    # ink = 7
                    # paper = 0
                    # ch = ' '
                    char_attrib = [
                        self.fgColorValues[7],  # fg
                        self.bgColorValues[0],  # bg
                        False,  # bold
                        False,  # italics
                        False,  # underscore
                        False,  # strikethrough
                        False,  # reverse
                        False,  # blink
                        "default",  # fontsize
                        "default",
                    ]  # fontfamily
                    line_attrib.append(char_attrib)
                    line_text += " "
                    continue
                ch = None
                try:
                    (sh,) = unpack("=H", data)
                    attr = (sh >> 8) & 0xFF
                    ch = sh & 0xFF
                    try:
                        if sh & self.hichar:
                            ch |= 0x100
                    except Exception as e:
                        self.env["runtime"]["DebugManager"].write_debug_out(
                            "vcsaDriver auto_decode_vcsa: Error processing character: "
                            + str(e),
                            debug.DebugLevel.ERROR,
                        )
                        ch = None
                    if self.hichar == 0x100:
                        attr >>= 1
                    ink = attr & 0x0F
                    paper = (attr >> 4) & 0x0F
                    if attr & 1:
                        blink = 1
                    # blink seems to be set always, ignore for now
                    blink = 0
                    bold = 0
                    if attr & 16:
                        bold = 1
                    # if (ink != 7) or (paper != 0):
                    #    print(ink,paper)
                except Exception as e:
                    self.env["runtime"]["DebugManager"].write_debug_out(
                        "vcsaDriver auto_decode_vcsa: Error processing attributes: "
                        + str(e),
                        debug.DebugLevel.ERROR,
                    )
                try:
                    line_text += self.charmap[ch]
                except KeyError:
                    line_text += "?"

                char_attrib = [
                    self.fgColorValues[ink],
                    self.bgColorValues[paper],
                    bold == 1,  # bold
                    False,  # italics
                    False,  # underscore
                    False,  # strikethrough
                    False,  # reverse
                    blink == 1,  # blink
                    "default",  # fontsize
                    "default",
                ]  # fontfamily
                line_attrib.append(char_attrib)
            all_text += line_text
            if y + 1 < rows:
                all_text += "\n"
            all_attrib.append(line_attrib)
        return str(all_text), all_attrib

    def get_screen_process_for_tty(self, tty_num):
        """Find the screen process associated with specific TTY"""
        try:
            result = subprocess.run([
                'ps', '-eo', 'pid,ppid,comm,tty,stat', '--no-headers'
            ], capture_output=True, text=True, timeout=2)
            
            for line in result.stdout.strip().split('\n'):
                if not line.strip():
                    continue
                parts = line.split()
                if len(parts) >= 4 and parts[2] == 'screen' and f'tty{tty_num}' in parts[3]:
                    return parts[0]  # Return PID of screen process
        except (subprocess.TimeoutExpired, subprocess.CalledProcessError):
            pass
        return None

    def get_screen_session_process(self, screen_tty_pid):
        """Get the session manager process for a TTY screen process"""
        try:
            result = subprocess.run([
                'ps', '-eo', 'pid,ppid,comm', '--no-headers'
            ], capture_output=True, text=True, timeout=2)
            
            for line in result.stdout.strip().split('\n'):
                if not line.strip():
                    continue
                parts = line.split()
                if len(parts) >= 3 and parts[1] == screen_tty_pid and parts[2] == 'screen':
                    return parts[0]  # Return session manager PID
        except (subprocess.TimeoutExpired, subprocess.CalledProcessError):
            pass
        return None

    def parse_active_app_from_pstree(self, pstree_output):
        """Parse pstree output to find currently active application"""
        try:
            # Look for processes that indicate active applications
            # Example: screen(1786)---bash(1787)---irssi(2016)
            import re
            
            # Find all application processes (non-bash, non-screen)
            # Pattern excludes --- prefix from pstree connection lines
            app_pattern = r'([a-zA-Z0-9_]+)\((\d+)\)'
            matches = re.findall(app_pattern, pstree_output)
            
            skip_processes = {'screen', 'bash', 'sh', 'grep', 'ps', 'sudo', 'sleep', 'clipboard_sync'}
            applications = []
            
            for app_name, pid in matches:
                if app_name.lower() not in skip_processes:
                    # Check if this process is in foreground state
                    try:
                        ps_result = subprocess.run([
                            'ps', '-p', pid, '-o', 'stat', '--no-headers'
                        ], capture_output=True, text=True, timeout=1)
                        
                        if ps_result.returncode == 0:
                            stat = ps_result.stdout.strip()
                            # Prioritize active processes
                            priority = 0
                            if 'S' in stat or 'R' in stat:  # Active processes
                                priority += 10
                            if '+' in stat:  # Foreground processes
                                priority += 20
                            if 'l' in stat.lower():  # Locked processes
                                priority += 5
                                
                            applications.append((app_name, pid, priority, stat))
                    except:
                        # If we can't check status, still consider it with low priority
                        applications.append((app_name, pid, 1, 'unknown'))
            
            # Sort by priority and return the highest priority application
            if applications:
                applications.sort(key=lambda x: (x[2], int(x[1])), reverse=True)
                best_app = applications[0][0].upper()
                
                self.env["runtime"]["DebugManager"].write_debug_out(
                    f"parse_active_app_from_pstree found {len(applications)} apps, selected: {best_app}",
                    debug.DebugLevel.INFO
                )
                return best_app
                
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                f"Error parsing pstree output: {str(e)}", debug.DebugLevel.ERROR
            )
        return None

    def get_app_from_screen_session(self, tty_num):
        """Get current application from active screen window on specific TTY"""
        try:
            # Find the screen session PID for this TTY
            ps_result = subprocess.run([
                'ps', '-eo', 'pid,ppid,comm,tty,args', '--no-headers'
            ], capture_output=True, text=True, timeout=2)
            
            if ps_result.returncode == 0:
                # Find the main screen process for our TTY
                screen_pid = None
                for line in ps_result.stdout.split('\n'):
                    if not line.strip():
                        continue
                    parts = line.split(None, 4)
                    if len(parts) >= 4 and parts[2] == 'screen' and f'tty{tty_num}' in parts[3]:
                        screen_pid = parts[0]
                        break
                
                if not screen_pid:
                    return None
                
                # Get the session manager screen process (child of TTY screen)
                session_pid = None
                for line in ps_result.stdout.split('\n'):
                    if not line.strip():
                        continue
                    parts = line.split(None, 4)
                    if len(parts) >= 3 and parts[1] == screen_pid and parts[2] == 'screen':
                        session_pid = parts[0]
                        break
                
                if not session_pid:
                    return None
                    
                self.env["runtime"]["DebugManager"].write_debug_out(
                    f"Found screen session PID: {session_pid}",
                    debug.DebugLevel.INFO
                )
                
                # Get all bash processes under this screen session
                bash_processes = []
                for line in ps_result.stdout.split('\n'):
                    if not line.strip():
                        continue
                    parts = line.split(None, 4)
                    if len(parts) >= 3 and parts[1] == session_pid and parts[2] == 'bash':
                        bash_processes.append(parts[0])
                
                # Check each bash for child applications, prioritizing active ones
                best_app = None
                best_priority = 0
                
                for bash_pid in bash_processes:
                    child_result = subprocess.run([
                        'ps', '--ppid', bash_pid, '-o', 'comm,stat', '--no-headers'
                    ], capture_output=True, text=True, timeout=1)
                    
                    if child_result.returncode == 0:
                        for child_line in child_result.stdout.split('\n'):
                            if not child_line.strip():
                                continue
                            child_parts = child_line.split()
                            if len(child_parts) >= 2:
                                child_comm, child_stat = child_parts[0], child_parts[1]
                                if child_comm.lower() not in ['grep', 'ps', 'bash', 'sh', 'sudo', 'sleep']:
                                    # Calculate priority based on process state
                                    priority = 0
                                    if 'R' in child_stat:  # Running
                                        priority += 30
                                    elif 'S' in child_stat:  # Sleeping
                                        priority += 20
                                    if '+' in child_stat:  # Foreground
                                        priority += 20
                                    
                                    self.env["runtime"]["DebugManager"].write_debug_out(
                                        f"Found child process: {child_comm} ({child_stat}) priority: {priority}",
                                        debug.DebugLevel.INFO
                                    )
                                    
                                    if priority > best_priority:
                                        best_app = child_comm.upper()
                                        best_priority = priority
                
                if best_app:
                    self.env["runtime"]["DebugManager"].write_debug_out(
                        f"Selected best app: {best_app} (priority: {best_priority})",
                        debug.DebugLevel.INFO
                    )
                    return best_app
                
        except (subprocess.TimeoutExpired, subprocess.CalledProcessError, FileNotFoundError) as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                f"Error getting app from screen session: {str(e)}", debug.DebugLevel.ERROR
            )
        return None

    def parse_latest_active_app_from_pstree(self, pstree_output):
        """Parse pstree output to find the most recently active application"""
        try:
            import re
            
            # Find all application processes with PIDs
            app_pattern = r'([a-zA-Z0-9_]+)\((\d+)\)'
            matches = re.findall(app_pattern, pstree_output)
            
            skip_processes = {'screen', 'bash', 'sh', 'grep', 'ps', 'sudo', 'sleep', 'clipboard_sync'}
            applications = []
            
            for app_name, pid in matches:
                if app_name.lower() not in skip_processes:
                    # Check if this process is still active and get its start time
                    try:
                        ps_result = subprocess.run([
                            'ps', '-p', pid, '-o', 'stat,lstart', '--no-headers'
                        ], capture_output=True, text=True, timeout=1)
                        
                        if ps_result.returncode == 0:
                            stat_info = ps_result.stdout.strip().split(None, 1)
                            if len(stat_info) >= 2:
                                stat = stat_info[0]
                                start_time = stat_info[1]
                                
                                # Prioritize processes that are active or have recent activity
                                priority = 0
                                if 'S' in stat or 'R' in stat:  # Running or sleeping (active)
                                    priority += 10
                                if '+' in stat:  # Foreground process
                                    priority += 20
                                    
                                applications.append((app_name, pid, priority, start_time))
                    except:
                        # If we can't get status, still consider it but with low priority
                        applications.append((app_name, pid, 1, 'unknown'))
            
            # Sort by priority (highest first), then by PID (most recent)
            if applications:
                applications.sort(key=lambda x: (x[2], int(x[1])), reverse=True)
                best_app = applications[0][0].upper()
                
                self.env["runtime"]["DebugManager"].write_debug_out(
                    f"Found {len(applications)} applications, selected: {best_app}",
                    debug.DebugLevel.INFO
                )
                return best_app
                
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                f"Error parsing latest active app from pstree: {str(e)}", 
                debug.DebugLevel.ERROR
            )
        return None

    def get_app_from_tmux_session(self, tty_num):
        """Get current application from tmux session on specific TTY"""
        try:
            # Try tmux list-panes to find active application
            result = subprocess.run([
                'tmux', 'list-panes', '-F', '#{pane_active} #{pane_current_command} #{pane_tty}'
            ], capture_output=True, text=True, timeout=2)
            
            if result.returncode == 0:
                for line in result.stdout.strip().split('\n'):
                    if not line.strip():
                        continue
                    parts = line.split()
                    if len(parts) >= 3 and parts[0] == '1':  # Active pane
                        tty_part = parts[2]
                        if tty_num in tty_part:
                            app = parts[1].upper()
                            if app not in ['BASH', 'SH']:
                                self.env["runtime"]["DebugManager"].write_debug_out(
                                    f"Found tmux application: {app}",
                                    debug.DebugLevel.INFO
                                )
                                return app
                    
        except (subprocess.TimeoutExpired, subprocess.CalledProcessError, FileNotFoundError):
            pass
        return None

    def get_curr_application(self):
        """Enhanced application detection supporting screen/tmux sessions.
        
        Multi-method approach:
        1. Try screen session detection via process tree analysis
        2. Try tmux session detection via tmux commands  
        3. Fall back to standard ps-based detection
        
        Updates:
            env['screen']['new_application']: Name of current application
            
        Features:
            - Detects applications inside screen/tmux sessions
            - Handles multiple screen sessions on different TTYs
            - Provides detailed debug logging for troubleshooting
        """
        curr_screen = self.env["screen"]["newTTY"]
        detected_app = None
        
        self.env["runtime"]["DebugManager"].write_debug_out(
            f"Starting application detection for TTY{curr_screen}",
            debug.DebugLevel.INFO
        )
        
        # Method 1: Try screen session detection
        try:
            detected_app = self.get_app_from_screen_session(curr_screen)
            if detected_app:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    f"Screen session detection found: {detected_app}",
                    debug.DebugLevel.INFO
                )
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                f"Screen session detection failed: {str(e)}", debug.DebugLevel.ERROR
            )
        
        # Method 2: Try tmux session detection
        if not detected_app:
            try:
                detected_app = self.get_app_from_tmux_session(curr_screen)
                if detected_app:
                    self.env["runtime"]["DebugManager"].write_debug_out(
                        f"Tmux session detection found: {detected_app}",
                        debug.DebugLevel.INFO
                    )
            except Exception as e:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    f"Tmux session detection failed: {str(e)}", debug.DebugLevel.ERROR
                )
        
        # Method 3: Fall back to standard ps-based detection
        if not detected_app:
            try:
                detected_app = self.get_app_via_standard_ps(curr_screen)
                if detected_app:
                    self.env["runtime"]["DebugManager"].write_debug_out(
                        f"Standard ps detection found: {detected_app}",
                        debug.DebugLevel.INFO
                    )
            except Exception as e:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    f"Standard ps detection failed: {str(e)}", debug.DebugLevel.ERROR
                )
        
        # Update application if we found one and it's different
        if detected_app and self.env["screen"]["new_application"] != detected_app:
            self.env["screen"]["new_application"] = detected_app
            self.env["runtime"]["DebugManager"].write_debug_out(
                f"Application changed to: {detected_app}",
                debug.DebugLevel.INFO
            )

    def get_app_via_standard_ps(self, curr_screen):
        """Fallback ps-based application detection for non-screen/tmux environments"""
        try:
            # Simple TTY-specific detection as fallback
            result = subprocess.run([
                'ps', '-t', f'tty{curr_screen}', '-o', 'comm,stat', '--no-headers'
            ], capture_output=True, text=True, timeout=2)
            
            if result.returncode == 0:
                for line in result.stdout.split('\n'):
                    if not line.strip():
                        continue
                    parts = line.split()
                    if len(parts) >= 2:
                        comm, stat = parts[0], parts[1]
                        if '+' in stat and comm.upper() not in ['BASH', 'SH', 'SCREEN', 'GREP', 'PS']:
                            self.env["runtime"]["DebugManager"].write_debug_out(
                                f"Fallback detection found: {comm}",
                                debug.DebugLevel.INFO
                            )
                            return comm.upper()
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                f"Standard ps detection error: {str(e)}", debug.DebugLevel.ERROR
            )
        return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.


from fenrirscreenreader.core.i18n import _


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return _("Line number for cursor")

    def run(self):
        cursor_pos = self.env["runtime"][
            "CursorManager"
        ].get_review_or_text_cursor()
        self.env["runtime"]["OutputManager"].present_text(
            str(cursor_pos["y"] + 1), interrupt=True
        )
        self.env["runtime"]["OutputManager"].announce_active_cursor()
        self.env["runtime"]["OutputManager"].present_text(
            " line number", interrupt=False
        )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core.i18n import _


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return _(
            "enables or disables automatic reading of new text as it appears"
        )

    def run(self):
        self.env["runtime"]["SettingsManager"].set_setting(
            "speech",
            "auto_read_incoming",
            str(
                not self.env["runtime"]["SettingsManager"].get_setting_as_bool(
                    "speech", "auto_read_incoming"
                )
            ),
        )
        if self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "speech", "auto_read_incoming"
        ):
            self.env["runtime"]["OutputManager"].present_text(
                _("autoread enabled"), sound_icon="", interrupt=True
            )
        else:
            self.env["runtime"]["OutputManager"].present_text(
                _("autoread disabled"), sound_icon="", interrupt=True
            )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import difflib
import re
import time
from fenrirscreenreader.core.i18n import _


class command:
    def __init__(self):
        self._update_times = []
        self._line_count_times = []
        self._batched_text = []
        self._last_flush_time = 0
        self._in_flood_mode = False
        self._pending_focus_text = ""
        self._pending_focus_time = 0

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return _("Announces incoming text changes")

    def _reset_flood_state(self):
        self._update_times = []
        self._line_count_times = []
        self._batched_text = []
        self._last_flush_time = 0
        self._in_flood_mode = False

    def _is_rapid_updates(self):
        current_time = time.time()
        window = self.env["runtime"]["SettingsManager"].get_setting_as_float(
            "speech", "rapid_update_window"
        )
        threshold = self.env["runtime"]["SettingsManager"].get_setting_as_int(
            "speech", "rapid_update_threshold"
        )

        self._update_times = [
            ts for ts in self._update_times if current_time - ts < window
        ]
        self._update_times.append(current_time)

        return len(self._update_times) >= threshold

    def _is_high_volume(self, delta_text):
        current_time = time.time()
        window = self.env["runtime"]["SettingsManager"].get_setting_as_float(
            "speech", "rapid_update_window"
        )
        threshold = self.env["runtime"]["SettingsManager"].get_setting_as_int(
            "speech", "flood_line_threshold"
        )

        line_count = max(1, delta_text.count("\n") + 1)
        self._line_count_times = [
            (ts, count)
            for ts, count in self._line_count_times
            if current_time - ts < window
        ]
        self._line_count_times.append((current_time, line_count))

        total_lines = sum(count for _, count in self._line_count_times)
        return total_lines >= threshold

    def _add_to_batch(self, text):
        new_lines = text.splitlines()
        if text.endswith("\n"):
            new_lines.append("")
        self._batched_text.extend(new_lines)

        max_lines = self.env["runtime"]["SettingsManager"].get_setting_as_int(
            "speech", "max_batch_lines"
        )
        if len(self._batched_text) > max_lines:
            self._batched_text = self._batched_text[-max_lines:]

    def _flush_batch(self):
        if not self._batched_text:
            return

        text = "\n".join(self._batched_text)
        self._batched_text = []
        self._last_flush_time = time.time()

        self.env["runtime"]["OutputManager"].present_text(
            text, interrupt=False, flush=False
        )

    def _was_handled_by_tab_completion(self, delta_text):
        """Check if this delta was already handled by tab completion to avoid duplicates"""
        if "tabCompletion" not in self.env["commandBuffer"]:
            return False

        tab_state = self.env["commandBuffer"]["tabCompletion"]

        # Check if this exact delta was processed recently by tab completion
        if (tab_state.get("lastProcessedDelta") == delta_text and
            tab_state.get("lastProcessedTime")):

            # Only suppress if processed within the last 50ms to avoid stale suppression
            # Reduced from 100ms to minimize false positives with rapid multi-line updates
            time_since_processed = time.time() - tab_state["lastProcessedTime"]
            if time_since_processed <= 0.05:
                return True

        return False

    def _normalize_line(self, line):
        return " ".join(line.split())

    def _is_focus_control_line(self, text):
        stripped = self._normalize_line(text)
        if not stripped:
            return False
        if re.fullmatch(r"<\s*[A-Za-z][A-Za-z0-9 _.-]{0,30}\s*>", stripped):
            return True
        return stripped.lower() in {
            "ok",
            "cancel",
            "yes",
            "no",
            "retry",
            "abort",
            "ignore",
            "continue",
        }

    def _is_transient_focus_only_update(self, delta_text):
        delta_lines = [
            self._normalize_line(line)
            for line in delta_text.splitlines()
            if line.strip()
        ]
        if len(delta_lines) != 1:
            return False
        if not self._is_focus_control_line(delta_lines[0]):
            return False

        screen_lines = [
            self._normalize_line(line)
            for line in self.env["screen"]["new_content_text"].split("\n")
            if line.strip()
        ]
        return screen_lines == delta_lines

    def _store_pending_focus_text(self, delta_text):
        delta_lines = [
            self._normalize_line(line)
            for line in delta_text.splitlines()
            if line.strip()
        ]
        self._pending_focus_text = "\n".join(delta_lines)
        self._pending_focus_time = time.time()

    def _take_pending_focus_text(self):
        if not self._pending_focus_text:
            return ""
        if time.time() - self._pending_focus_time > 0.5:
            self._pending_focus_text = ""
            self._pending_focus_time = 0
            return ""
        pending = self._pending_focus_text
        self._pending_focus_text = ""
        self._pending_focus_time = 0
        return pending

    def _is_subsequence(self, subset_lines, source_lines):
        subset_index = 0
        for source_line in source_lines:
            if source_line == subset_lines[subset_index]:
                subset_index += 1
                if subset_index == len(subset_lines):
                    return True
        return False

    def _prefer_inserted_lower_screen_text(self, delta_text):
        delta_lines = [
            line for line in delta_text.splitlines() if line.strip() != ""
        ]
        if len(delta_lines) < 2:
            return delta_text

        old_lines = self.env["screen"]["old_content_text"].split("\n")
        new_lines = self.env["screen"]["new_content_text"].split("\n")
        if len(old_lines) != len(new_lines) or len(new_lines) < 4:
            return delta_text

        normalized_old_lines = [
            self._normalize_line(line) for line in old_lines
        ]
        normalized_new_lines = [
            self._normalize_line(line) for line in new_lines
        ]
        matcher = difflib.SequenceMatcher(
            None, normalized_old_lines, normalized_new_lines, autojunk=False
        )

        top_lines_changed = False
        inserted_lines = []
        lower_screen_start = max(2, len(new_lines) // 2)

        for tag, old_start, old_end, new_start, new_end in matcher.get_opcodes():
            if tag in {"replace", "delete"} and old_start < 2:
                top_lines_changed = True
            if tag != "insert" or new_start < lower_screen_start:
                continue
            inserted_lines.extend(
                line for line in new_lines[new_start:new_end] if line.strip() != ""
            )

        if not top_lines_changed or not inserted_lines:
            return delta_text
        if all(
            self._is_focus_control_line(line) for line in inserted_lines
        ):
            return delta_text

        normalized_delta_lines = [
            self._normalize_line(line) for line in delta_lines
        ]
        normalized_inserted_lines = [
            self._normalize_line(line) for line in inserted_lines
        ]
        if not self._is_subsequence(
            normalized_inserted_lines, normalized_delta_lines
        ):
            return delta_text
        if normalized_delta_lines == normalized_inserted_lines:
            return delta_text

        return "\n".join(inserted_lines)

    def run(self):
        if not self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "speech", "auto_read_incoming"
        ):
            return
        # is there something to read?
        if not self.env["runtime"]["ScreenManager"].is_delta(ignoreSpace=True):
            return
        if self.env["screen"].get("new_delta_is_typing", False):
            return

        delta_text = self.env["screen"]["new_delta"]

        if self.env["runtime"]["ScreenManager"].is_screen_change():
            self._reset_flood_state()
            self._pending_focus_text = ""
            self._pending_focus_time = 0
        
        # Skip if tab completion already handled this delta
        if self._was_handled_by_tab_completion(delta_text):
            return

        # this must be a keyecho or something
        # if len(self.env['screen']['new_delta'].strip(' \n\t')) <= 1:
        x_move = abs(
            self.env["screen"]["new_cursor"]["x"]
            - self.env["screen"]["old_cursor"]["x"]
        )
        y_move = abs(
            self.env["screen"]["new_cursor"]["y"]
            - self.env["screen"]["old_cursor"]["y"]
        )

        if (x_move >= 1) and x_move == len(delta_text):
            # if len(self.env['screen']['new_delta'].strip(' \n\t0123456789'))
            # <= 2:
            if "\n" not in delta_text:
                return

        if self._is_transient_focus_only_update(delta_text):
            self._store_pending_focus_text(delta_text)
            return

        rapid = self._is_rapid_updates()
        high_volume = self._is_high_volume(delta_text)

        if (rapid and high_volume) or self._in_flood_mode:
            if not self._in_flood_mode:
                self._last_flush_time = time.time()
                self._in_flood_mode = True

            self._add_to_batch(delta_text)

            interval = self.env["runtime"][
                "SettingsManager"
            ].get_setting_as_float("speech", "batch_flush_interval")
            if time.time() - self._last_flush_time >= interval:
                self._flush_batch()

            if not rapid or not high_volume:
                if self._batched_text:
                    self._flush_batch()
                self._in_flood_mode = False
            return

        # print(x_move, y_move, len(self.env['screen']['new_delta']), len(self.env['screen']['newNegativeDelta']))
        delta_text = self._prefer_inserted_lower_screen_text(delta_text)
        pending_focus_text = self._take_pending_focus_text()
        if pending_focus_text:
            delta_text = "\n".join([delta_text, pending_focus_text])
        self.env["runtime"]["OutputManager"].present_text(
            delta_text, interrupt=False, flush=False
        )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core.i18n import _
from fenrirscreenreader.core import debug


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return _("enables or disables tracking of highlighted text")

    def run(self):
        highlight_mode = self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "focus", "highlight"
        )
        cursor_mode = self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "focus", "cursor"
        )
        table_mode = self.env["runtime"]["TableManager"].is_table_mode()
        
        self.env["runtime"]["DebugManager"].write_debug_out(
            f"toggle_highlight_tracking: highlight={highlight_mode}, "
            f"cursor={cursor_mode}, table={table_mode}",
            debug.DebugLevel.INFO
        )

        # Cycle through modes: highlight → cursor → table → highlight
        if highlight_mode and not table_mode:
            # Switch to cursor mode
            self.env["runtime"]["SettingsManager"].set_setting(
                "focus", "highlight", "False"
            )
            self.env["runtime"]["SettingsManager"].set_setting(
                "focus", "cursor", "True"
            )
            self.env["runtime"]["TableManager"].set_table_mode(False)
            self.env["runtime"]["OutputManager"].present_text(
                _("cursor tracking"), sound_icon="", interrupt=True
            )
        elif cursor_mode and not table_mode:
            # Switch to table mode
            self.env["runtime"]["SettingsManager"].set_setting(
                "focus", "highlight", "False"
            )
            self.env["runtime"]["SettingsManager"].set_setting(
                "focus", "cursor", "False"
            )
            self.env["runtime"]["TableManager"].set_table_mode(True)
            self.env["runtime"]["OutputManager"].present_text(
                _("table mode"), sound_icon="", interrupt=True
            )
        else:
            # Switch to highlight mode (default) - also handles stuck table mode
            self.env["runtime"]["SettingsManager"].set_setting(
                "focus", "highlight", "True"
            )
            self.env["runtime"]["SettingsManager"].set_setting(
                "focus", "cursor", "False"
            )
            self.env["runtime"]["TableManager"].set_table_mode(False)
            self.env["runtime"]["OutputManager"].present_text(
                _("highlight tracking"), sound_icon="", interrupt=True
            )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader import fenrirVersion
from fenrirscreenreader.core.i18n import _


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return _("Present the version of Fenrir currently in use.")

    def run(self):
        try:
            self.env["runtime"]["OutputManager"].present_text(
                f"Fenrir screen reader version "
                f"{fenrirVersion.version}-{fenrirVersion.code_name}",
                interrupt=True,
            )
        except Exception as e:
            self.env["runtime"]["OutputManager"].present_text(
                _("Version information is unavailable."), interrupt=True
            )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
"""
Fenrir PEP8 Validation and Auto-Fix Tool

Validates Python code style using flake8 and applies safe automatic fixes.
Designed to work with Fenrir's existing codebase while respecting timing-critical code.

Usage:
    python3 tools/validate_pep8.py                    # Check all Python files
    python3 tools/validate_pep8.py --fix-safe        # Auto-fix safe issues
    python3 tools/validate_pep8.py --check-only      # Exit with error if issues found
    python3 tools/validate_pep8.py --staged          # Check only staged files
"""

import os
import sys
import argparse
import subprocess
import tempfile
from pathlib import Path


class PEP8Validator:
    def __init__(self, verbose=True):
        self.verbose = verbose
        self.errors = []
        self.warnings = []
        self.fixes_applied = []
        
    def log(self, message, level="INFO"):
        """Log a message with appropriate formatting."""
        if not self.verbose and level == "INFO":
            return
            
        colors = {
            "INFO": "\033[0;36m",     # Cyan
            "SUCCESS": "\033[0;32m",  # Green
            "WARNING": "\033[1;33m",  # Yellow
            "ERROR": "\033[0;31m",    # Red
            "HEADER": "\033[1;34m",   # Bold Blue
        }
        
        reset = "\033[0m"
        color = colors.get(level, "")
        
        if level == "HEADER":
            print(f"\n{color}{'='*60}")
            print(f"{message}")
            print(f"{'='*60}{reset}")
        else:
            symbol = {
                "SUCCESS": "✓",
                "ERROR": "✗", 
                "WARNING": "⚠",
                "INFO": "•"
            }.get(level, "•")
            
            print(f"{color}{symbol} {message}{reset}")
    
    def check_flake8_available(self):
        """Check if flake8 is available."""
        try:
            result = subprocess.run(["flake8", "--version"], 
                                  capture_output=True, text=True, timeout=5)
            if result.returncode == 0:
                version = result.stdout.strip().split('\n')[0]
                self.log(f"Using flake8: {version}")
                return True
            else:
                return False
        except (subprocess.TimeoutExpired, FileNotFoundError):
            return False
    
    def get_python_files(self, directory=None, staged_only=False):
        """Get list of Python files to check."""
        if staged_only:
            try:
                result = subprocess.run([
                    "git", "diff", "--cached", "--name-only", "--diff-filter=ACM"
                ], capture_output=True, text=True, timeout=10)
                
                if result.returncode == 0:
                    files = [f for f in result.stdout.strip().split('\n') 
                            if f.endswith('.py') and Path(f).exists()]
                    return [Path(f) for f in files if f]
                else:
                    self.warnings.append("Could not get staged files, checking all files")
                    staged_only = False
            except subprocess.TimeoutExpired:
                self.warnings.append("Git command timed out, checking all files")
                staged_only = False
        
        if not staged_only:
            directory = Path(directory or "src/fenrirscreenreader")
            if not directory.exists():
                self.errors.append(f"Directory {directory} does not exist")
                return []
            
            python_files = list(directory.rglob("*.py"))
            # Filter out cache and build directories
            python_files = [f for f in python_files if not any(
                part.startswith(('__pycache__', '.git', 'build', 'dist')) 
                for part in f.parts)]
            return python_files
        
        return []
    
    def run_flake8(self, files, select=None, ignore=None):
        """Run flake8 on the given files."""
        if not files:
            return True, ""
        
        cmd = ["flake8"]
        
        if select:
            cmd.extend(["--select", select])
        if ignore:
            cmd.extend(["--ignore", ignore])
        
        # Add files
        cmd.extend([str(f) for f in files])
        
        try:
            result = subprocess.run(cmd, capture_output=True, text=True, timeout=30)
            return result.returncode == 0, result.stdout
        except subprocess.TimeoutExpired:
            self.errors.append("flake8 command timed out")
            return False, ""
        except Exception as e:
            self.errors.append(f"Failed to run flake8: {e}")
            return False, ""
    
    def categorize_issues(self, flake8_output):
        """Categorize flake8 issues by severity and safety for auto-fixing."""
        lines = flake8_output.strip().split('\n')
        issues = {'critical': [], 'safe_fixable': [], 'manual': []}
        
        for line in lines:
            if not line.strip():
                continue
                
            # Parse flake8 output: filename:line:col: code message
            parts = line.split(':', 3)
            if len(parts) < 4:
                continue
                
            filename = parts[0]
            line_num = parts[1]
            col = parts[2]
            code_msg = parts[3].strip()
            
            code = code_msg.split()[0] if code_msg else ""
            
            # Categorize by error code
            if code.startswith('E9') or code.startswith('F'):
                # Critical syntax/import errors
                issues['critical'].append(line)
            elif code in ['E111', 'E114', 'E117', 'E121', 'E122', 'E123', 'E124', 
                         'E125', 'E126', 'E127', 'E128', 'E129', 'E131', 'E133',
                         'W291', 'W292', 'W293']:
                # Safe indentation and whitespace fixes
                # But skip timing-critical files
                if not any(critical in filename.lower() for critical in 
                          ['evdevdriver', 'vcsadriver', 'screenmanager', 'inputmanager']):
                    issues['safe_fixable'].append(line)
                else:
                    issues['manual'].append(line)
            else:
                # Everything else needs manual review
                issues['manual'].append(line)
        
        return issues
    
    def apply_safe_fixes(self, files):
        """Apply safe automatic fixes using autopep8."""
        try:
            # Check if autopep8 is available
            result = subprocess.run(["autopep8", "--version"], 
                                  capture_output=True, text=True, timeout=5)
            if result.returncode != 0:
                self.warnings.append("autopep8 not available for auto-fixing")
                return False
        except (subprocess.TimeoutExpired, FileNotFoundError):
            self.warnings.append("autopep8 not available for auto-fixing")
            return False
        
        fixed_count = 0
        
        for file_path in files:
            # Skip timing-critical files
            if any(critical in str(file_path).lower() for critical in 
                  ['evdevdriver', 'vcsadriver', 'screenmanager', 'inputmanager']):
                self.log(f"Skipping timing-critical file: {file_path}", "WARNING")
                continue
            
            try:
                # Apply safe fixes only
                cmd = [
                    "autopep8", 
                    "--in-place",
                    "--select", "E111,E114,E117,E121,E122,E123,E124,E125,E126,E127,E128,E129,E131,E133,W291,W292,W293",
                    str(file_path)
                ]
                
                result = subprocess.run(cmd, capture_output=True, text=True, timeout=10)
                if result.returncode == 0:
                    self.fixes_applied.append(f"Applied safe PEP8 fixes to {file_path}")
                    fixed_count += 1
                else:
                    self.warnings.append(f"Could not auto-fix {file_path}: {result.stderr}")
                    
            except subprocess.TimeoutExpired:
                self.warnings.append(f"Auto-fix timed out for {file_path}")
            except Exception as e:
                self.warnings.append(f"Error auto-fixing {file_path}: {e}")
        
        return fixed_count > 0
    
    def validate_files(self, files, fix_safe=False):
        """Validate Python files for PEP8 compliance."""
        if not files:
            self.log("No Python files to validate")
            return True
        
        self.log(f"Validating {len(files)} Python files")
        
        # Run comprehensive flake8 check
        success, output = self.run_flake8(files)
        
        if not output.strip():
            self.log("All files pass PEP8 validation", "SUCCESS")
            return True
        
        # Categorize issues
        issues = self.categorize_issues(output)
        
        # Report critical issues
        if issues['critical']:
            self.log(f"Critical issues found ({len(issues['critical'])}):", "ERROR")
            for issue in issues['critical'][:10]:  # Limit output
                self.log(f"  {issue}", "ERROR")
            if len(issues['critical']) > 10:
                self.log(f"  ... and {len(issues['critical']) - 10} more", "ERROR")
        
        # Handle safe fixable issues
        if issues['safe_fixable']:
            if fix_safe:
                self.log(f"Auto-fixing {len(issues['safe_fixable'])} safe issues...", "INFO")
                # Get unique files from safe_fixable issues
                fix_files = set()
                for issue in issues['safe_fixable']:
                    filename = issue.split(':')[0]
                    fix_files.add(Path(filename))
                
                if self.apply_safe_fixes(fix_files):
                    self.log("Safe auto-fixes applied", "SUCCESS")
                    # Re-run flake8 to see remaining issues
                    success, output = self.run_flake8(files)
                    if output.strip():
                        remaining_issues = self.categorize_issues(output)
                        issues = remaining_issues
                    else:
                        issues = {'critical': [], 'safe_fixable': [], 'manual': []}
            else:
                self.log(f"Safe fixable issues found ({len(issues['safe_fixable'])}):", "WARNING")
                for issue in issues['safe_fixable'][:5]:
                    self.log(f"  {issue}", "WARNING")
                if len(issues['safe_fixable']) > 5:
                    self.log(f"  ... and {len(issues['safe_fixable']) - 5} more", "WARNING")
                self.log("Run with --fix-safe to auto-fix these", "INFO")
        
        # Report manual issues
        if issues['manual']:
            self.log(f"Manual review needed ({len(issues['manual'])}):", "WARNING")
            for issue in issues['manual'][:5]:
                self.log(f"  {issue}", "WARNING")
            if len(issues['manual']) > 5:
                self.log(f"  ... and {len(issues['manual']) - 5} more", "WARNING")
        
        # Return success if only manual issues remain (non-critical)
        return len(issues['critical']) == 0
    
    def generate_report(self):
        """Generate final validation report."""
        total_issues = len(self.errors) + len(self.warnings)
        
        if self.fixes_applied:
            self.log(f"\nAUTO-FIXES APPLIED ({len(self.fixes_applied)}):", "HEADER")
            for fix in self.fixes_applied:
                self.log(fix, "SUCCESS")
        
        if self.errors:
            self.log(f"\nERRORS ({len(self.errors)}):", "HEADER")
            for error in self.errors:
                self.log(error, "ERROR")
        
        if self.warnings:
            self.log(f"\nWARNINGS ({len(self.warnings)}):", "HEADER")
            for warning in self.warnings:
                self.log(warning, "WARNING")
        
        if len(self.errors) == 0:
            self.log("\n✅ PEP8 VALIDATION PASSED", "SUCCESS")
            if self.warnings:
                self.log("Non-critical style issues found - consider manual review", "INFO")
            return True
        else:
            self.log("\n❌ PEP8 VALIDATION FAILED", "ERROR")
            self.log("Critical issues must be fixed", "ERROR")
            return False


def main():
    parser = argparse.ArgumentParser(description='Validate and fix PEP8 compliance in Fenrir')
    parser.add_argument('--fix-safe', action='store_true',
                       help='Apply safe automatic fixes (avoids timing-critical files)')
    parser.add_argument('--check-only', action='store_true',
                       help='Exit with non-zero code if issues found')
    parser.add_argument('--staged', action='store_true',
                       help='Check only staged files')
    parser.add_argument('--quiet', action='store_true',
                       help='Reduce output verbosity')
    parser.add_argument('--directory', default='src/fenrirscreenreader',
                       help='Directory to scan (default: src/fenrirscreenreader)')
    
    args = parser.parse_args()
    
    validator = PEP8Validator(verbose=not args.quiet)
    
    validator.log("FENRIR PEP8 VALIDATION", "HEADER")
    
    # Check if flake8 is available
    if not validator.check_flake8_available():
        validator.log("flake8 is required but not available", "ERROR")
        validator.log("Install with: pip install flake8", "INFO")
        if args.fix_safe:
            validator.log("For auto-fixing, also install: pip install autopep8", "INFO")
        sys.exit(1)
    
    # Get files to validate
    files = validator.get_python_files(
        directory=args.directory if not args.staged else None,
        staged_only=args.staged
    )
    
    if not files:
        validator.log("No Python files found to validate")
        sys.exit(0)
    
    # Validate files
    success = validator.validate_files(files, fix_safe=args.fix_safe)
    
    # Generate report
    validation_passed = validator.generate_report()
    
    if args.check_only and not validation_passed:
        sys.exit(1)
    elif validation_passed:
        sys.exit(0)
    else:
        sys.exit(1)


if __name__ == '__main__':
    main()



# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import multiprocessing
import threading
import time
from ctypes import c_bool
from multiprocessing.sharedctypes import Value
from select import select

from fenrirscreenreader.core import debug
from fenrirscreenreader.core import inputData
from fenrirscreenreader.core.eventData import FenrirEventType
from fenrirscreenreader.core.inputDriver import InputDriver as inputDriver

_evdevAvailable = False
_udevAvailable = False
_evdevAvailableError = ""
_udevAvailableError = ""
try:
    import evdev
    from evdev import InputDevice
    from evdev import UInput
    from evdev import ecodes as e

    _evdevAvailable = True

except Exception as e:
    _evdevAvailableError = str(e)

try:
    import pyudev

    _udevAvailable = True
except Exception as e:
    _udevAvailableError = str(e)


class driver(inputDriver):
    """Linux evdev input driver for Fenrir screen reader.
    
    This driver provides access to Linux input devices through the evdev interface,
    allowing Fenrir to capture keyboard input, manage device grabbing for exclusive
    access, and inject synthetic input events.
    
    Features:
    - Automatic device detection and hotplug support via udev
    - Device grabbing to prevent input from reaching other applications
    - Key event mapping and filtering
    - UInput support for synthetic key injection
    - Multi-device support with thread-safe access
    
    Attributes:
        iDevices (dict): Map of file descriptor to InputDevice objects
        iDevicesFD (list): Shared list of file descriptors for multiprocessing
        uDevices (dict): Map of file descriptor to UInput devices
        gDevices (dict): Map of file descriptor to grab status
        iDeviceNo (int): Total number of input devices
        UInputinject (UInput): Device for injecting synthetic events
        _deviceLock (Lock): Thread lock for device access
    """
    def __init__(self):
        inputDriver.__init__(self)
        self._manager = multiprocessing.Manager()
        self.iDevices = {}
        self.iDevicesFD = self._manager.list()
        self.uDevices = {}
        self.gDevices = {}
        self.iDeviceNo = 0
        self.watch_dog = Value(c_bool, True)
        self.UInputinject = (
            UInput() if _evdevAvailable and "UInput" in globals() else None
        )
        self._deviceLock = threading.Lock()

    def initialize(self, environment):
        """Initialize the evdev input driver.
        
        Sets up device monitoring, starts watchdog threads for device hotplug
        detection and input monitoring, and configures the input subsystem.
        
        Args:
            environment: Fenrir environment dictionary with runtime managers
            
        Note:
            Requires evdev and optionally pyudev libraries. Falls back gracefully
            if libraries are not available.
        """
        self.env = environment
        global _evdevAvailable
        global _udevAvailable
        global _evdevAvailableError
        global _udevAvailableError
        if not _udevAvailable:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "InputDriver:" + _udevAvailableError, debug.DebugLevel.ERROR
            )
        if not _evdevAvailable:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "InputDriver:" + _evdevAvailableError, debug.DebugLevel.ERROR
            )
            return

        if _udevAvailable:
            self.env["runtime"]["ProcessManager"].add_custom_event_thread(
                self.plug_input_device_watchdog_udev
            )
        self.env["runtime"]["ProcessManager"].add_custom_event_thread(
            self.input_watchdog
        )
        self._initialized = True

    def plug_input_device_watchdog_udev(self, active, event_queue):
        """Monitor for input device hotplug events via udev.
        
        Runs in a separate thread to detect when input devices are
        plugged/unplugged and generates appropriate events for device
        management.
        
        Args:
            active: Shared boolean controlling the watchdog loop
            event_queue: Queue for sending device events to main process
            
        Events Generated:
            FenrirEventType.plug_input_device: When new devices are detected
            
        Note:
            Filters out virtual devices and devices from assistive technologies
            like BRLTTY to avoid conflicts.
        """
        context = pyudev.Context()
        monitor = pyudev.Monitor.from_netlink(context)
        monitor.filter_by(subsystem="input")
        monitor.start()
        ignore_plug = False
        while active.value:
            valid_devices = []
            device = monitor.poll(1)
            while device:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "plug_input_device_watchdog_udev:" + str(device),
                    debug.DebugLevel.INFO,
                )
                try:
                    try:
                        # FIX: Check if attributes exist before accessing them
                        if (
                            hasattr(device, "name")
                            and device.name
                            and device.name.upper()
                            in ["", "SPEAKUP", "FENRIR-UINPUT"]
                        ):
                            ignore_plug = True
                        if (
                            hasattr(device, "phys")
                            and device.phys
                            and device.phys.upper()
                            in ["", "SPEAKUP", "FENRIR-UINPUT"]
                        ):
                            ignore_plug = True
                        if (
                            hasattr(device, "name")
                            and device.name
                            and "BRLTTY" in device.name.upper()
                        ):
                            ignore_plug = True
                    except Exception as e:
                        self.env["runtime"]["DebugManager"].write_debug_out(
                            "plug_input_device_watchdog_udev CHECK NAME CRASH: "
                            + str(e),
                            debug.DebugLevel.ERROR,
                        )

                    if not ignore_plug:
                        virtual = (
                            "/sys/devices/virtual/input/" in device.sys_path
                        )
                        if device.device_node:
                            valid_devices.append(
                                {
                                    "device": device.device_node,
                                    "virtual": virtual,
                                }
                            )
                except Exception as e:
                    self.env["runtime"]["DebugManager"].write_debug_out(
                        "plug_input_device_watchdog_udev APPEND CRASH: "
                        + str(e),
                        debug.DebugLevel.ERROR,
                    )
                try:
                    poll_timeout = 1
                    device = monitor.poll(poll_timeout)
                except Exception:
                    device = None
                ignore_plug = False
            if valid_devices:
                event_queue.put(
                    {
                        "Type": FenrirEventType.plug_input_device,
                        "data": valid_devices,
                    }
                )
        return time.time()

    def input_watchdog(self, active, event_queue):
        """Main input monitoring loop.
        
        Monitors all registered input devices for key events using select().
        Processes key events, maps them to Fenrir's internal format, and
        forwards them to the main application.
        
        Args:
            active: Shared boolean controlling the watchdog loop  
            event_queue: Queue for sending input events to main process
            
        Events Generated:
            FenrirEventType.keyboard_input: For key press/release events
            
        Note:
            Uses thread-safe device access and handles device disconnection
            gracefully. Non-keyboard events are forwarded to UInput devices.
        """
        try:
            while active.value:
                # Get a snapshot of devices for select() to avoid lock
                # contention
                with self._deviceLock:
                    devices_snapshot = self.iDevices.copy()

                if not devices_snapshot:
                    time.sleep(0.1)
                    continue

                r, w, x = select(devices_snapshot, [], [], 0.8)
                event = None
                found_key_in_sequence = False
                foreward = False
                event_fired = False
                for fd in r:
                    # Check if device still exists before accessing
                    with self._deviceLock:
                        if fd not in self.iDevices:
                            continue
                        device = self.iDevices[fd]
                        udevice = self.uDevices.get(fd)

                    try:
                        event = device.read_one()
                    except Exception as e:
                        self.env["runtime"]["DebugManager"].write_debug_out(
                            "evdevDriver handle_input_event: Error reading event: "
                            + str(e),
                            debug.DebugLevel.ERROR,
                        )
                        self.remove_device(fd)
                        continue
                    while event:
                        self.env["runtime"]["DebugManager"].write_debug_out(
                            "input_watchdog: EVENT:" + str(event),
                            debug.DebugLevel.INFO,
                        )
                        self.env["input"]["event_buffer"].append(
                            [device, udevice, event]
                        )
                        if event.type == evdev.events.EV_KEY:
                            if not found_key_in_sequence:
                                found_key_in_sequence = True
                            if event.code != 0:
                                curr_map_event = self.map_event(event)
                                if not curr_map_event:
                                    event = device.read_one()
                                    continue
                                if not isinstance(
                                    curr_map_event["event_name"], str
                                ):
                                    event = device.read_one()
                                    continue
                                if curr_map_event["event_state"] in [0, 1, 2]:
                                    event_queue.put(
                                        {
                                            "Type": FenrirEventType.keyboard_input,
                                            "data": curr_map_event.copy(),
                                        }
                                    )
                                    event_fired = True
                        else:
                            # Forward non-keyboard events: 2=EV_REL, 3=EV_ABS, 17=EV_LED
                            if event.type in [2, 3, 17]:
                                foreward = True

                        event = device.read_one()
                    if not found_key_in_sequence:
                        if foreward and not event_fired:
                            self.write_event_buffer()
                            self.clear_event_buffer()
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "INPUT WATCHDOG CRASH: " + str(e), debug.DebugLevel.ERROR
            )

    def write_event_buffer(self):
        if not self._initialized:
            return
        for iDevice, uDevice, event in self.env["input"]["event_buffer"]:
            try:
                if uDevice:
                    if self.gDevices[iDevice.fd]:
                        self.write_u_input(uDevice, event)
            except Exception as e:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "evdevDriver write_event_buffer: Error writing event: "
                    + str(e),
                    debug.DebugLevel.ERROR,
                )

    def write_u_input(self, uDevice, event):
        if not self._initialized:
            return
        uDevice.write_event(event)
        time.sleep(0.0000002)
        uDevice.syn()

    def update_input_devices(self, new_devices=None, init=False):
        """Update the list of monitored input devices.
        
        Scans for available input devices and adds suitable ones based on
        the configured device mode. Supports filtering by device type and
        name matching.
        
        Args:
            new_devices (list, optional): Specific devices to add
            init (bool): Whether this is initial device setup
            
        Device Modes:
            - 'ALL': Monitor all keyboard devices
            - 'NOMICE': Monitor keyboards but exclude pointing devices
            - Device names: Comma-separated list of specific device names
            
        Note:
            Automatically filters out virtual devices, assistive technology
            devices, and devices with insufficient key counts.
        """
        if init:
            self.remove_all_devices()

        device_file_list = None

        if new_devices and not init:
            if not isinstance(new_devices, list):
                new_devices = [new_devices]
            device_file_list = new_devices
        else:
            device_file_list = evdev.list_devices()
            if len(device_file_list) == self.iDeviceNo:
                return
        if not device_file_list:
            return

        mode = (
            self.env["runtime"]["SettingsManager"]
            .get_setting("keyboard", "device")
            .upper()
        )

        i_devices_files = []
        for device in self.iDevices:
            i_devices_files.append(self.iDevices[device].path)

        event_type = evdev.events
        for deviceFile in device_file_list:
            try:
                if not deviceFile:
                    continue
                if deviceFile == "":
                    continue
                if deviceFile in i_devices_files:
                    continue
                try:
                    with open(deviceFile) as f:
                        pass
                except Exception as e:
                    self.env["runtime"]["DebugManager"].write_debug_out(
                        "evdevDriver update_input_devices: Error opening device file: "
                        + str(e),
                        debug.DebugLevel.ERROR,
                    )
                    continue
                # 3 pos absolute
                # 2 pos relative
                # 1 Keys
                try:
                    curr_device = evdev.InputDevice(deviceFile)
                except Exception as e:
                    self.env["runtime"]["DebugManager"].write_debug_out(
                        "evdevDriver createDeviceType: Error creating device: "
                        + str(e),
                        debug.DebugLevel.ERROR,
                    )
                    continue
                try:
                    # FIX: Check if attributes exist before accessing them
                    if (
                        hasattr(curr_device, "name")
                        and curr_device.name
                        and curr_device.name.upper()
                        in ["", "SPEAKUP", "FENRIR-UINPUT"]
                    ):
                        continue
                    if (
                        hasattr(curr_device, "phys")
                        and curr_device.phys
                        and curr_device.phys.upper()
                        in ["", "SPEAKUP", "FENRIR-UINPUT"]
                    ):
                        continue
                    if (
                        hasattr(curr_device, "name")
                        and curr_device.name
                        and "BRLTTY" in curr_device.name.upper()
                    ):
                        continue
                except Exception as e:
                    self.env["runtime"]["DebugManager"].write_debug_out(
                        "evdevDriver: Error checking device capabilities: "
                        + str(e),
                        debug.DebugLevel.ERROR,
                    )
                cap = curr_device.capabilities()
                if mode in ["ALL", "NOMICE"]:
                    if event_type.EV_KEY in cap:
                        if (
                            116 in cap[event_type.EV_KEY]
                            and len(cap[event_type.EV_KEY]) < 10
                        ):
                            self.env["runtime"][
                                "DebugManager"
                            ].write_debug_out(
                                "Device Skipped (has 116):" + curr_device.name,
                                debug.DebugLevel.INFO,
                            )
                            continue
                        # Check if device has numpad keys - use lower threshold for dedicated numpads
                        numpad_keys = [
                            evdev.ecodes.KEY_KP0, evdev.ecodes.KEY_KP1, evdev.ecodes.KEY_KP2,
                            evdev.ecodes.KEY_KP3, evdev.ecodes.KEY_KP4, evdev.ecodes.KEY_KP5,
                            evdev.ecodes.KEY_KP6, evdev.ecodes.KEY_KP7, evdev.ecodes.KEY_KP8,
                            evdev.ecodes.KEY_KP9, evdev.ecodes.KEY_KPPLUS, evdev.ecodes.KEY_KPMINUS,
                            evdev.ecodes.KEY_KPASTERISK, evdev.ecodes.KEY_KPSLASH, evdev.ecodes.KEY_KPENTER,
                            evdev.ecodes.KEY_KPDOT
                        ]
                        
                        has_numpad_keys = any(key in cap[event_type.EV_KEY] for key in numpad_keys)
                        min_key_threshold = 10 if has_numpad_keys else 60
                        
                        if len(cap[event_type.EV_KEY]) < min_key_threshold:
                            threshold_type = "numpad" if has_numpad_keys else "keyboard"
                            self.env["runtime"][
                                "DebugManager"
                            ].write_debug_out(
                                f"Device Skipped (< {min_key_threshold} keys for {threshold_type}):"
                                + curr_device.name,
                                debug.DebugLevel.INFO,
                            )
                            continue
                        if mode == "ALL":
                            self.add_device(curr_device)
                            self.env["runtime"][
                                "DebugManager"
                            ].write_debug_out(
                                "Device added (ALL):"
                                + self.iDevices[curr_device.fd].name,
                                debug.DebugLevel.INFO,
                            )
                        elif mode == "NOMICE":
                            if not (
                                (event_type.EV_REL in cap)
                                or (event_type.EV_ABS in cap)
                            ):
                                self.add_device(curr_device)
                                self.env["runtime"][
                                    "DebugManager"
                                ].write_debug_out(
                                    "Device added (NOMICE):"
                                    + self.iDevices[curr_device.fd].name,
                                    debug.DebugLevel.INFO,
                                )
                            else:
                                self.env["runtime"][
                                    "DebugManager"
                                ].write_debug_out(
                                    "Device Skipped (NOMICE):"
                                    + curr_device.name,
                                    debug.DebugLevel.INFO,
                                )
                    else:
                        self.env["runtime"]["DebugManager"].write_debug_out(
                            "Device Skipped (no EV_KEY):" + curr_device.name,
                            debug.DebugLevel.INFO,
                        )
                elif curr_device.name.upper() in mode.split(","):
                    self.add_device(curr_device)
                    self.env["runtime"]["DebugManager"].write_debug_out(
                        "Device added (Name):"
                        + self.iDevices[curr_device.fd].name,
                        debug.DebugLevel.INFO,
                    )
            except Exception as e:
                try:
                    device_name = (
                        curr_device.name
                        if hasattr(curr_device, "name")
                        else "unknown"
                    )
                    self.env["runtime"]["DebugManager"].write_debug_out(
                        "Device Skipped (Exception): "
                        + deviceFile
                        + " "
                        + device_name
                        + " "
                        + str(e),
                        debug.DebugLevel.INFO,
                    )
                except Exception as ex:
                    self.env["runtime"]["DebugManager"].write_debug_out(
                        "Device Skipped (Exception): "
                        + deviceFile
                        + " "
                        + str(ex),
                        debug.DebugLevel.INFO,
                    )
        self.iDeviceNo = len(evdev.list_devices())
        self.update_m_pi_devices_fd()

    def update_m_pi_devices_fd(self):
        try:
            for fd in self.iDevices:
                if fd not in self.iDevicesFD:
                    self.iDevicesFD.append(fd)
            for fd in self.iDevicesFD:
                if fd not in self.iDevices:
                    self.iDevicesFD.remove(fd)
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "evdevDriver update_m_pi_devices_fd: Error updating device file descriptors: "
                + str(e),
                debug.DebugLevel.ERROR,
            )

    def map_event(self, event):
        if not self._initialized:
            return None
        if not event:
            return None
        m_event = inputData.input_event
        try:
            # mute is a list = ['KEY_MIN_INTERESTING', 'KEY_MUTE']
            m_event["event_name"] = evdev.ecodes.keys[event.code]
            if isinstance(m_event["event_name"], list):
                if len(m_event["event_name"]) > 0:
                    m_event["event_name"] = m_event["event_name"][0]
                    if isinstance(m_event["event_name"], list):
                        if len(m_event["event_name"]) > 0:
                            m_event["event_name"] = m_event["event_name"][0]
            m_event["event_value"] = event.code
            m_event["event_sec"] = event.sec
            m_event["event_usec"] = event.usec
            m_event["event_state"] = event.value
            m_event["event_type"] = event.type
            return m_event
        except Exception as e:
            return None

    def get_led_state(self, led=0):
        if not self.has_i_devices():
            return False
        # 0 = Numlock
        # 1 = Capslock
        # 2 = Rollen
        
        # Use the first device with LED capability as authoritative source
        # to avoid inconsistent readings from multiple devices during initialization
        for fd, dev in self.iDevices.items():
            # Check if device has LED capability (capability 17)
            if 17 in dev.capabilities():
                return led in dev.leds()
        
        # Fallback to old behavior if no device has LED capability
        for fd, dev in self.iDevices.items():
            if led in dev.leds():
                return True
        return False

    def toggle_led_state(self, led=0):
        if not self.has_i_devices():
            return False
        led_state = self.get_led_state(led)
        for i in self.iDevices:
            if self.gDevices[i]:
                # 17 LEDs
                if 17 in self.iDevices[i].capabilities():
                    if led_state == 1:
                        self.iDevices[i].set_led(led, 0)
                    else:
                        self.iDevices[i].set_led(led, 1)

    def grab_all_devices(self):
        if not self._initialized:
            return True
        ok = True
        for fd in self.iDevices:
            if not self.gDevices[fd]:
                ok = ok and self.grab_device(fd)
        return ok

    def ungrab_all_devices(self):
        if not self._initialized:
            return True
        ok = True
        for fd in self.iDevices:
            if self.gDevices[fd]:
                ok = ok and self.ungrab_device(fd)
        return ok

    def create_u_input_dev(self, fd):
        if not self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "keyboard", "grab_devices"
        ):
            self.uDevices[fd] = None
            return
        try:
            test = self.uDevices[fd]
            return
        except KeyError:
            self.uDevices[fd] = None
        if self.uDevices[fd] is not None:
            return
        if "UInput" not in globals():
            return
        try:
            self.uDevices[fd] = UInput.from_device(
                self.iDevices[fd], name="fenrir-uinput", phys="fenrir-uinput"
            )
        except Exception as e:
            try:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "InputDriver evdev: compat fallback:  " + str(e),
                    debug.DebugLevel.WARNING,
                )
                dev = self.iDevices[fd]
                cap = dev.capabilities()
                del cap[0]
                self.uDevices[fd] = UInput(
                    cap, name="fenrir-uinput", phys="fenrir-uinput"
                )
            except Exception as e:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "InputDriver evdev: init Uinput not possible:  " + str(e),
                    debug.DebugLevel.ERROR,
                )
                return

    def add_device(self, newDevice):
        """Add a new input device to the monitoring list.
        
        Creates the necessary data structures for device monitoring,
        sets up UInput forwarding if device grabbing is enabled, and
        initializes device state.
        
        Args:
            newDevice: evdev.InputDevice object to add
            
        Note:
            Thread-safe operation. Automatically cleans up on failure.
        """
        self.env["runtime"]["DebugManager"].write_debug_out(
            "InputDriver evdev: device added:  "
            + str(newDevice.fd)
            + " "
            + str(newDevice),
            debug.DebugLevel.INFO,
        )
        with self._deviceLock:
            try:
                self.iDevices[newDevice.fd] = newDevice
                self.create_u_input_dev(newDevice.fd)
                self.gDevices[newDevice.fd] = False
            except Exception as e:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "InputDriver evdev: error adding device: " + str(e),
                    debug.DebugLevel.ERROR,
                )
                # if it doesnt work clean up
                try:
                    del self.iDevices[newDevice.fd]
                except Exception as e:
                    self.env["runtime"]["DebugManager"].write_debug_out(
                        "evdevDriver remove_device: Error removing iDevice: "
                        + str(e),
                        debug.DebugLevel.ERROR,
                    )
                try:
                    del self.uDevices[newDevice.fd]
                except Exception as e:
                    self.env["runtime"]["DebugManager"].write_debug_out(
                        "evdevDriver remove_device: Error removing uDevice: "
                        + str(e),
                        debug.DebugLevel.ERROR,
                    )
                try:
                    del self.gDevices[newDevice.fd]
                except Exception as e:
                    self.env["runtime"]["DebugManager"].write_debug_out(
                        "evdevDriver add_device: Error cleaning up gDevice: "
                        + str(e),
                        debug.DebugLevel.ERROR,
                    )

    def grab_device(self, fd):
        """Grab exclusive access to an input device.
        
        Takes exclusive control of the device, preventing other applications
        from receiving its input. Also resets modifier key states to prevent
        stuck keys.
        
        Args:
            fd (int): File descriptor of device to grab
            
        Returns:
            bool: True if grab successful, False otherwise
            
        Note:
            Only effective if grabDevices setting is enabled.
        """
        if not self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "keyboard", "grab_devices"
        ):
            return True

        # FIX: Handle exception variable scope correctly
        grab_error = None
        try:
            self.iDevices[fd].grab()
            self.gDevices[fd] = True
            self.env["runtime"]["DebugManager"].write_debug_out(
                "InputDriver evdev: grab device ("
                + str(self.iDevices[fd].name)
                + ")",
                debug.DebugLevel.INFO,
            )
            # Reset modifier keys on successful grab
            if self.uDevices[fd]:
                modifier_keys = [
                    e.KEY_LEFTCTRL,
                    e.KEY_RIGHTCTRL,
                    e.KEY_LEFTALT,
                    e.KEY_RIGHTALT,
                    e.KEY_LEFTSHIFT,
                    e.KEY_RIGHTSHIFT,
                ]
                for key in modifier_keys:
                    try:
                        self.uDevices[fd].write(e.EV_KEY, key, 0)  # 0 = key up
                        self.uDevices[fd].syn()
                    except Exception as mod_error:
                        self.env["runtime"]["DebugManager"].write_debug_out(
                            "Failed to reset modifier key: " + str(mod_error),
                            debug.DebugLevel.WARNING,
                        )
        except IOError:
            if not self.gDevices[fd]:
                return False
        except Exception as ex:
            grab_error = ex

        if grab_error:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "InputDriver evdev: grabing not possible: " + str(grab_error),
                debug.DebugLevel.ERROR,
            )
            return False

        return True

    def ungrab_device(self, fd):
        """Release exclusive access to an input device.
        
        Returns control of the device to the system, allowing other
        applications to receive its input.
        
        Args:
            fd (int): File descriptor of device to ungrab
            
        Returns:
            bool: True if ungrab successful, False otherwise
        """
        if not self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "keyboard", "grab_devices"
        ):
            return True

        # FIX: Handle exception variable scope correctly
        ungrab_error = None
        try:
            self.iDevices[fd].ungrab()
            self.gDevices[fd] = False
            self.env["runtime"]["DebugManager"].write_debug_out(
                "InputDriver evdev: ungrab device ("
                + str(self.iDevices[fd].name)
                + ")",
                debug.DebugLevel.INFO,
            )
        except IOError:
            if self.gDevices[fd]:
                return False
        except Exception as ex:
            ungrab_error = ex

        if ungrab_error:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "InputDriver evdev: ungrabing not possible: "
                + str(ungrab_error),
                debug.DebugLevel.ERROR,
            )
            return False

        return True

    def remove_device(self, fd):
        """Remove an input device from monitoring.
        
        Cleanly removes a device by ungrabbing it, closing file handles,
        and cleaning up all associated data structures.
        
        Args:
            fd (int): File descriptor of device to remove
            
        Note:
            Thread-safe operation with comprehensive error handling.
        """
        with self._deviceLock:
            try:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "InputDriver evdev: device removed:  "
                    + str(fd)
                    + " "
                    + str(self.iDevices[fd]),
                    debug.DebugLevel.INFO,
                )
            except Exception as e:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "InputDriver evdev: device removed:  "
                    + str(fd)
                    + " Error: "
                    + str(e),
                    debug.DebugLevel.INFO,
                )
            self.clear_event_buffer()
            try:
                self.ungrab_device(fd)
            except Exception as e:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "evdevDriver remove_device: Error ungrabbing device "
                    + str(fd)
                    + ": "
                    + str(e),
                    debug.DebugLevel.ERROR,
                )
            try:
                self.iDevices[fd].close()
            except Exception as e:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "evdevDriver remove_device: Error closing iDevice "
                    + str(fd)
                    + ": "
                    + str(e),
                    debug.DebugLevel.ERROR,
                )
            try:
                self.uDevices[fd].close()
            except Exception as e:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "evdevDriver remove_device: Error closing uDevice "
                    + str(fd)
                    + ": "
                    + str(e),
                    debug.DebugLevel.ERROR,
                )
            try:
                del self.iDevices[fd]
            except Exception as e:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "evdevDriver remove_device: Error deleting iDevice "
                    + str(fd)
                    + ": "
                    + str(e),
                    debug.DebugLevel.ERROR,
                )
            try:
                del self.uDevices[fd]
            except Exception as e:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "evdevDriver remove_device: Error deleting uDevice "
                    + str(fd)
                    + ": "
                    + str(e),
                    debug.DebugLevel.ERROR,
                )
            try:
                del self.gDevices[fd]
            except Exception as e:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "evdevDriver remove_device: Error deleting gDevice "
                    + str(fd)
                    + ": "
                    + str(e),
                    debug.DebugLevel.ERROR,
                )
            self.update_m_pi_devices_fd()

    def has_i_devices(self):
        if not self._initialized:
            return False
        if not self.iDevices:
            return False
        if len(self.iDevices) == 0:
            return False
        return True

    def send_key(self, key, state):
        """Inject a synthetic key event.
        
        Sends a key press or release event using UInput. Used for
        features like key forwarding and macro execution.
        
        Args:
            key (str): Key name (e.g., 'KEY_A')
            state (int): Key state (0=release, 1=press, 2=repeat)
        """
        if not self._initialized:
            return
        if self.UInputinject is None:
            return
        try:
            self.UInputinject.write(e.EV_KEY, e.ecodes[key], state)
            self.UInputinject.syn()
        except Exception as ex:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "evdevDriver send_key: Error sending key "
                + str(key)
                + ": "
                + str(ex),
                debug.DebugLevel.ERROR,
            )

    def remove_all_devices(self):
        if not self.has_i_devices():
            return
        devices = self.iDevices.copy()
        for fd in devices:
            self.remove_device(fd)
        self.iDevices.clear()
        self.uDevices.clear()
        self.gDevices.clear()
        self.iDeviceNo = 0



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.


# this command is just to initialize stuff.
# like init index lists in MemoryManager
# it is not useful to execute it

from fenrirscreenreader.core.i18n import _


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment
        # clipboard
        self.env["runtime"]["MemoryManager"].add_index_list(
            "clipboardHistory",
            self.env["runtime"]["SettingsManager"].get_setting_as_int(
                "general", "number_of_clipboards"
            ),
        )

    def shutdown(self):
        pass

    def get_description(self):
        return "No description found"

    def run(self):
        pass

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import datetime

from fenrirscreenreader.core.i18n import _


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return _("Test mc search functionality")

    def run(self):
        # Test command for mc search operations
        test_message = _("MC search test: This demonstrates search functionality")
        
        # present the test message
        self.env["runtime"]["OutputManager"].present_text(
            test_message, sound_icon="", interrupt=True
        )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core import debug


class CursorManager:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def should_process_numpad_commands(self):
        """
        Check if numpad commands should be processed based on numlock state
        Return True if numlock is OFF (commands should work)
        Return False if numlock is ON (let keys type numbers)
        """
        # Return False if numlock is ON
        return not self.env["input"]["new_num_lock"]

    def shutdown(self):
        pass

    def clear_marks(self):
        self.env["commandBuffer"]["Marks"]["1"] = None
        self.env["commandBuffer"]["Marks"]["2"] = None

    def is_mark_set(self):
        return self.env["commandBuffer"]["Marks"]["1"] is not None

    def is_single_mark(self):
        return (
            self.env["commandBuffer"]["Marks"]["1"] is not None
            and self.env["commandBuffer"]["Marks"]["2"] is None
        )

    def is_multible_mark(self):
        return (
            self.env["commandBuffer"]["Marks"]["1"] is not None
            and self.env["commandBuffer"]["Marks"]["2"] is not None
        )

    def set_mark(self):
        curr_cursor = None
        if self.env["screen"]["newCursorReview"]:
            curr_cursor = self.env["screen"]["newCursorReview"].copy()
        else:
            curr_cursor = self.env["screen"]["new_cursor"].copy()
        if not self.env["commandBuffer"]["Marks"]["1"]:
            self.env["commandBuffer"]["Marks"]["1"] = curr_cursor.copy()
            return 1
        else:
            self.env["commandBuffer"]["Marks"]["2"] = curr_cursor.copy()
            return 2
        return 0

    def get_review_or_text_cursor(self):
        if self.env["screen"]["newCursorReview"]:
            return self.env["screen"]["newCursorReview"].copy()
        else:
            return self.env["screen"]["new_cursor"].copy()

    def clear_review_cursor(self):
        if not self.is_review_mode():
            return
        self.env["screen"]["oldCursorReview"] = None
        self.env["screen"]["newCursorReview"] = None

    def is_cursor_horizontal_move(self):
        return (
            self.env["screen"]["new_cursor"]["x"]
            != self.env["screen"]["old_cursor"]["x"]
        )

    def is_cursor_vertical_move(self):
        return (
            self.env["screen"]["new_cursor"]["y"]
            != self.env["screen"]["old_cursor"]["y"]
        )

    def is_review_mode(self):
        return self.env["screen"]["newCursorReview"] is not None

    def enter_review_mode_curr_text_cursor(self, overwrite=False):
        if self.is_review_mode() and not overwrite:
            return
        self.env["screen"]["oldCursorReview"] = self.env["screen"][
            "newCursorReview"
        ]
        if not self.env["screen"]["newCursorReview"]:
            self.env["screen"]["newCursorReview"] = self.env["screen"][
                "new_cursor"
            ].copy()
            if (
                self.env["runtime"]["SettingsManager"].get_setting_as_bool(
                    "focus", "highlight"
                )
                and self.env["runtime"][
                    "AttributeManager"
                ].is_attribute_cursor_active()
            ):
                self.env["screen"]["newCursorReview"] = (
                    self.env["runtime"]["AttributeManager"]
                    .get_curr_attribute_cursor()
                    .copy()
                )

    def set_review_cursor_position(self, x, y):
        if not self.is_review_mode():
            self.enter_review_mode_curr_text_cursor()
        self.env["screen"]["oldCursorReview"] = self.env["screen"][
            "newCursorReview"
        ]
        self.env["screen"]["newCursorReview"]["x"] = x
        self.env["screen"]["newCursorReview"]["y"] = y

    def is_application_window_set(self):
        try:
            curr_app = self.env["runtime"][
                "ApplicationManager"
            ].get_current_application()
            if (
                curr_app in self.env["commandBuffer"]["windowArea"]
                and "1" in self.env["commandBuffer"]["windowArea"][curr_app]
                and self.env["commandBuffer"]["windowArea"][curr_app]["1"]
                is not None
            ):
                return True
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "CursorManager is_application_window_set: Error checking window area: "
                + str(e),
                debug.DebugLevel.ERROR,
            )
        return False

    def set_window_for_application(self, start=None, end=None):
        x1 = 0
        x2 = 0
        y1 = 0
        y2 = 0
        if start is None:
            if not self.env["commandBuffer"]["Marks"]["1"]:
                return False
            else:
                x1 = self.env["commandBuffer"]["Marks"]["1"]["x"]
                y1 = self.env["commandBuffer"]["Marks"]["1"]["y"]
        else:
            x1 = start["x"]
            y1 = start["y"]
        if end is None:
            if not self.env["commandBuffer"]["Marks"]["2"]:
                return False
            else:
                x2 = self.env["commandBuffer"]["Marks"]["2"]["x"]
                y2 = self.env["commandBuffer"]["Marks"]["2"]["y"]
        else:
            x2 = end["x"]
            y2 = end["y"]

        curr_app = self.env["runtime"][
            "ApplicationManager"
        ].get_current_application()
        self.env["commandBuffer"]["windowArea"][curr_app] = {}

        if x1 * y1 <= x2 * y2:
            self.env["commandBuffer"]["windowArea"][curr_app]["1"] = {
                "x": x1,
                "y": y1,
            }
            self.env["commandBuffer"]["windowArea"][curr_app]["2"] = {
                "x": x2,
                "y": y2,
            }
        else:
            self.env["commandBuffer"]["windowArea"][curr_app]["1"] = {
                "x": x2,
                "y": y2,
            }
            self.env["commandBuffer"]["windowArea"][curr_app]["2"] = {
                "x": x1,
                "y": y1,
            }
        return True

    def clear_window_for_application(self):
        curr_app = self.env["runtime"][
            "ApplicationManager"
        ].get_current_application()
        try:
            del self.env["commandBuffer"]["windowArea"][curr_app]
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "CursorManager clear_window_for_application: Error clearing window area: "
                + str(e),
                debug.DebugLevel.ERROR,
            )
            return False
        return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core.i18n import _
from fenrirscreenreader.core import debug


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return _("Read all text line by line from current position")

    def run(self):
        self.env["runtime"]["OutputManager"].present_text(
            "Read-all functionality temporarily disabled",
            interrupt=True
        )
        return
        # Check if speechd is available for callbacks
        try:
            speech_driver = self.env['runtime']['SpeechDriver']
            if not (hasattr(speech_driver, '_sd') and speech_driver._sd):
                self.env["runtime"]["OutputManager"].present_text(
                    _("Read all requires speech-dispatcher - not available with current speech driver"),
                    interrupt=True
                )
                return
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                f"ReadAllByLine run: Error checking speech driver: {e}",
                debug.DebugLevel.ERROR
            )
            return
        
        # Check if ReadAllManager is available
        if "ReadAllManager" not in self.env["runtime"]:
            self.env["runtime"]["OutputManager"].present_text(
                _("Read all manager not in runtime"),
                interrupt=True
            )
            return
        elif not self.env["runtime"]["ReadAllManager"]:
            self.env["runtime"]["OutputManager"].present_text(
                _("Read all manager is None"),
                interrupt=True
            )
            return
        
        # Start continuous line-by-line reading
        if self.env["runtime"]["ReadAllManager"].is_active():
            # Stop if already active
            self.env["runtime"]["ReadAllManager"].stop_read_all()
        else:
            # Start line-by-line reading
            self.env["runtime"]["ReadAllManager"].start_read_all('line')


    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core.i18n import _
from fenrirscreenreader.utils import screen_utils


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return _("Reads attributes of current cursor position")

    def run(self):
        # is it enabled?
        if not self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "general", "hasattributes"
        ):
            return
        # is a vertical change?
        if not (
            self.env["runtime"]["CursorManager"].is_cursor_vertical_move()
            or self.env["runtime"]["CursorManager"].is_cursor_horizontal_move()
        ):
            return

        cursor_pos = self.env["screen"]["new_cursor"]

        if not self.env["runtime"]["AttributeManager"].has_attributes(
            cursor_pos
        ):
            return
        self.env["runtime"]["OutputManager"].present_text(
            "has attribute", sound_icon="HasAttributes", interrupt=False
        )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import os
import sys
import time

from fenrirscreenreader.core import debug
from fenrirscreenreader.core.eventData import FenrirEventType
from fenrirscreenreader.core.inputDriver import InputDriver as inputDriver

try:
    from Xlib import X
    from Xlib import XK
    from Xlib import display
    from Xlib.error import BadAccess
    from Xlib.error import BadWindow
except Exception as x_error:
    X = None
    XK = None
    display = None
    BadAccess = Exception
    BadWindow = Exception
    _x_error = x_error
else:
    _x_error = None


def build_keysym_name_map():
    if XK is None:
        return {}
    names = {}
    for attr_name in dir(XK):
        if not attr_name.startswith("XK_"):
            continue
        value = getattr(XK, attr_name)
        if not isinstance(value, int):
            continue
        names.setdefault(value, attr_name[3:])
    return names


KEYSYM_NAME_MAP = build_keysym_name_map()


class X11DriverError(RuntimeError):
    pass


class driver(inputDriver):
    """X11 terminal-scoped keyboard input driver."""

    key_name_overrides = {
        "BackSpace": "KEY_BACKSPACE",
        "Return": "KEY_ENTER",
        "ISO_Enter": "KEY_ENTER",
        "Tab": "KEY_TAB",
        "Escape": "KEY_ESC",
        "space": "KEY_SPACE",
        "minus": "KEY_MINUS",
        "underscore": "KEY_MINUS",
        "_": "KEY_MINUS",
        "equal": "KEY_EQUAL",
        "plus": "KEY_EQUAL",
        "+": "KEY_EQUAL",
        "bracketleft": "KEY_LEFTBRACE",
        "bracketright": "KEY_RIGHTBRACE",
        "backslash": "KEY_BACKSLASH",
        "semicolon": "KEY_SEMICOLON",
        "apostrophe": "KEY_APOSTROPHE",
        "grave": "KEY_GRAVE",
        "comma": "KEY_COMMA",
        "period": "KEY_DOT",
        "slash": "KEY_SLASH",
        "Shift_L": "KEY_LEFTSHIFT",
        "Shift_R": "KEY_RIGHTSHIFT",
        "Control_L": "KEY_LEFTCTRL",
        "Control_R": "KEY_RIGHTCTRL",
        "Alt_L": "KEY_LEFTALT",
        "Alt_R": "KEY_RIGHTALT",
        "Meta_L": "KEY_LEFTMETA",
        "Meta_R": "KEY_RIGHTMETA",
        "Super_L": "KEY_LEFTMETA",
        "Super_R": "KEY_RIGHTMETA",
        "Multi_key": "KEY_COMPOSE",
        "Caps_Lock": "KEY_CAPSLOCK",
        "Num_Lock": "KEY_NUMLOCK",
        "Scroll_Lock": "KEY_SCROLLLOCK",
        "Insert": "KEY_INSERT",
        "Delete": "KEY_DELETE",
        "Home": "KEY_HOME",
        "End": "KEY_END",
        "Prior": "KEY_PAGEUP",
        "Next": "KEY_PAGEDOWN",
        "Page_Up": "KEY_PAGEUP",
        "Page_Down": "KEY_PAGEDOWN",
        "Up": "KEY_UP",
        "Down": "KEY_DOWN",
        "Left": "KEY_LEFT",
        "Right": "KEY_RIGHT",
        "KP_0": "KEY_KP0",
        "KP_Insert": "KEY_KP0",
        "KP_1": "KEY_KP1",
        "KP_End": "KEY_KP1",
        "KP_2": "KEY_KP2",
        "KP_Down": "KEY_KP2",
        "KP_3": "KEY_KP3",
        "KP_Page_Down": "KEY_KP3",
        "KP_4": "KEY_KP4",
        "KP_Left": "KEY_KP4",
        "KP_5": "KEY_KP5",
        "KP_Begin": "KEY_KP5",
        "KP_6": "KEY_KP6",
        "KP_Right": "KEY_KP6",
        "KP_7": "KEY_KP7",
        "KP_Home": "KEY_KP7",
        "KP_8": "KEY_KP8",
        "KP_Up": "KEY_KP8",
        "KP_9": "KEY_KP9",
        "KP_Page_Up": "KEY_KP9",
        "KP_Decimal": "KEY_KPDOT",
        "KP_Delete": "KEY_KPDOT",
        "KP_Add": "KEY_KPPLUS",
        "KP_Subtract": "KEY_KPMINUS",
        "KP_Multiply": "KEY_KPASTERISK",
        "KP_Divide": "KEY_KPSLASH",
        "KP_Enter": "KEY_KPENTER",
        "KP_Equal": "KEY_KPEQUAL",
    }

    modifier_masks = {
        "KEY_SHIFT": X.ShiftMask if X else 1,
        "KEY_LEFTSHIFT": X.ShiftMask if X else 1,
        "KEY_RIGHTSHIFT": X.ShiftMask if X else 1,
        "KEY_CTRL": X.ControlMask if X else 4,
        "KEY_LEFTCTRL": X.ControlMask if X else 4,
        "KEY_RIGHTCTRL": X.ControlMask if X else 4,
        "KEY_ALT": X.Mod1Mask if X else 8,
        "KEY_LEFTALT": X.Mod1Mask if X else 8,
        "KEY_RIGHTALT": X.Mod1Mask if X else 8,
        "KEY_META": X.Mod4Mask if X else 64,
        "KEY_LEFTMETA": X.Mod4Mask if X else 64,
        "KEY_RIGHTMETA": X.Mod4Mask if X else 64,
    }

    modifier_key_names = set(modifier_masks.keys())

    def __init__(self):
        inputDriver.__init__(self)
        self.display = None
        self.root = None
        self.window = None
        self.window_id = None
        self.active = True
        self.num_lock_mask = 0
        self.grabbed = set()
        self.grab_signature = None
        self.interesting_keys = set()
        self.fenrir_keys = set()
        self.failed_grabs = 0
        self.modifier_state = 0

    def initialize(self, environment):
        self.env = environment
        if display is None:
            self.fail_startup("python-xlib is not available: " + str(_x_error))
        self.display = display.Display()
        self.root = self.display.screen().root
        self.window_id = self.resolve_window_id()
        if self.window_id is None:
            self.fail_startup(
                "No X11 target window found. Use --x11-window-id or launch from an X11 terminal."
            )
        self.write_debug(
            "x11Driver target window "
            + self.format_window_id(self.window_id)
            + ", active window "
            + self.format_window_id(self.get_active_window_id())
            + ", WINDOWID="
            + os.environ.get("WINDOWID", ""),
            debug.DebugLevel.INFO,
        )
        self.window = self.display.create_resource_object("window", self.window_id)
        self.window.change_attributes(
            event_mask=(
                X.KeyPressMask
                | X.KeyReleaseMask
                | X.FocusChangeMask
                | X.StructureNotifyMask
            )
        )
        self.num_lock_mask = self.find_num_lock_mask()
        self.refresh_modifier_state()
        self.refresh_interesting_keys()
        self.refresh_grabs(force=True)
        self.env["runtime"]["ProcessManager"].add_custom_event_thread(
            self.input_watchdog
        )
        self._initialized = True

    def shutdown(self):
        self.ungrab_all_devices()
        try:
            if self.display:
                self.display.close()
        except Exception:
            pass
        self._initialized = False

    def fail_startup(self, message):
        print("Fenrir X11 driver error: " + message, file=sys.stderr)
        raise SystemExit(1)

    def resolve_window_id(self):
        configured_window = self.env["runtime"]["SettingsManager"].get_setting(
            "keyboard", "x11_window_id"
        )
        for candidate in [configured_window, os.environ.get("WINDOWID", "")]:
            window_id = self.parse_window_id(candidate)
            if window_id is not None:
                return window_id
        return self.get_active_window_id()

    def parse_window_id(self, value):
        if value is None:
            return None
        value = str(value).strip()
        if value == "":
            return None
        try:
            return int(value, 0)
        except ValueError:
            raise X11DriverError("Invalid X11 window id: " + value)

    def get_active_window_id(self):
        atom = self.display.intern_atom("_NET_ACTIVE_WINDOW")
        prop = self.root.get_full_property(atom, X.AnyPropertyType)
        if prop is None or len(prop.value) == 0:
            return None
        return int(prop.value[0])

    def format_window_id(self, window_id):
        if window_id is None:
            return "None"
        return hex(int(window_id))

    def find_num_lock_mask(self):
        num_lock_keysym = XK.string_to_keysym("Num_Lock")
        if not num_lock_keysym:
            return 0
        num_lock_keycode = self.display.keysym_to_keycode(num_lock_keysym)
        if not num_lock_keycode:
            return 0
        modifier_map = self.display.get_modifier_mapping()
        masks = [
            X.ShiftMask,
            X.LockMask,
            X.ControlMask,
            X.Mod1Mask,
            X.Mod2Mask,
            X.Mod3Mask,
            X.Mod4Mask,
            X.Mod5Mask,
        ]
        for index, keycodes in enumerate(modifier_map):
            if num_lock_keycode in keycodes:
                return masks[index]
        return 0

    def input_watchdog(self, active, event_queue):
        while active.value:
            try:
                self.refresh_grabs()
                if not self.display.pending_events():
                    time.sleep(0.01)
                    continue
                event = self.display.next_event()
                self.handle_x_event(event, event_queue)
            except BadWindow:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "x11Driver target window disappeared",
                    debug.DebugLevel.ERROR,
                )
                break
            except Exception as e:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "x11Driver input watchdog error: " + str(e),
                    debug.DebugLevel.ERROR,
                )
                time.sleep(0.1)

    def handle_x_event(self, event, event_queue):
        event_type = getattr(event, "type", None)
        if event_type == X.FocusIn:
            self.active = True
            self.clear_event_buffer()
            return
        if event_type == X.FocusOut:
            self.active = False
            self.reset_input_state()
            return
        if event_type not in [X.KeyPress, X.KeyRelease]:
            return
        if not self.active and not self.event_is_in_target_tree(event):
            return
        input_event = self.map_event(event)
        if input_event is None:
            return
        key_name = input_event["event_name"]
        if not self.should_emit_key(key_name):
            return
        self.update_modifier_state_from_event(input_event)
        self.write_debug(
            "x11Driver key event "
            + key_name
            + " state "
            + str(input_event["event_state"])
            + " raw state "
            + str(getattr(event, "state", 0)),
            debug.DebugLevel.INFO,
        )
        self.env["input"]["event_buffer"].append(input_event.copy())
        event_queue.put(
            {
                "Type": FenrirEventType.keyboard_input,
                "data": input_event,
            }
        )

    def event_is_in_target_tree(self, event):
        event_window = getattr(event, "event", None)
        if event_window is None:
            return False
        window_id = getattr(event_window, "id", None)
        if window_id is None:
            return False
        return window_id == self.window_id or self.window_has_ancestor(
            window_id, self.window_id
        )

    def window_has_ancestor(self, window_id, ancestor_id):
        try:
            window = self.display.create_resource_object("window", window_id)
            while window.id != self.root.id:
                if window.id == ancestor_id:
                    return True
                parent = window.query_tree().parent
                if parent is None or parent.id == window.id:
                    return False
                window = parent
        except Exception:
            return False
        return False

    def map_event(self, event):
        key_name = self.keycode_to_key_name(event.detail, event.state)
        if key_name is None:
            return None
        event_time = time.time()
        return {
            "event_name": key_name,
            "event_value": event.detail,
            "event_sec": int(event_time),
            "event_usec": int((event_time % 1) * 1000000),
            "event_state": 1 if event.type == X.KeyPress else 0,
            "event_type": event.type,
            "event_raw_state": getattr(event, "state", 0),
        }

    def refresh_modifier_state(self):
        try:
            pointer = self.root.query_pointer()
            self.modifier_state = getattr(pointer, "mask", 0)
        except Exception:
            self.modifier_state = 0

    def update_modifier_state_from_event(self, input_event):
        key_name = input_event["event_name"]
        event_state = input_event["event_state"]
        raw_state = input_event.get("event_raw_state", self.modifier_state)
        self.modifier_state = raw_state
        if key_name == "KEY_NUMLOCK":
            self.update_lock_modifier_state(self.num_lock_mask, event_state)
        elif key_name == "KEY_CAPSLOCK":
            self.update_lock_modifier_state(X.LockMask, event_state)
        else:
            modifier_mask = self.modifier_masks.get(key_name, 0)
            if modifier_mask:
                if event_state == 1:
                    self.modifier_state |= modifier_mask
                elif event_state == 0:
                    self.modifier_state &= ~modifier_mask

    def update_lock_modifier_state(self, modifier_mask, event_state):
        if not modifier_mask or event_state != 1:
            return
        if self.modifier_state & modifier_mask:
            self.modifier_state &= ~modifier_mask
        else:
            self.modifier_state |= modifier_mask

    def keycode_to_key_name(self, keycode, state=0):
        key_names = []
        for keysym_name in self.keycode_to_keysym_names(keycode):
            key_name = self.keysym_name_to_key_name(keysym_name)
            if key_name is None or key_name in key_names:
                continue
            key_names.append(key_name)
        for key_name in key_names:
            if key_name.startswith("KEY_KP"):
                return key_name
        if key_names:
            return key_names[0]
        return None

    def keycode_to_keysym_names(self, keycode):
        names = []
        for column in range(8):
            keysym = self.display.keycode_to_keysym(keycode, column)
            keysym_name = self.keysym_to_name(keysym)
            if keysym_name and keysym_name not in names:
                names.append(keysym_name)
        return names

    def keysym_to_name(self, keysym):
        keysym_name = KEYSYM_NAME_MAP.get(keysym)
        if keysym_name and keysym >= 0xFF00:
            return keysym_name
        keysym_string = XK.keysym_to_string(keysym)
        if keysym_string:
            return keysym_string
        return keysym_name

    def keysym_name_to_key_name(self, keysym_name):
        if not keysym_name:
            return None
        if keysym_name in self.key_name_overrides:
            return self.key_name_overrides[keysym_name]
        if len(keysym_name) == 1:
            if keysym_name.isalpha():
                return "KEY_" + keysym_name.upper()
            if keysym_name.isdigit():
                return "KEY_" + keysym_name
        if keysym_name.startswith("F") and keysym_name[1:].isdigit():
            return "KEY_" + keysym_name
        return None

    def should_emit_key(self, key_name):
        return key_name is not None

    def refresh_interesting_keys(self):
        input_manager = self.env["runtime"]["InputManager"]
        self.fenrir_keys = set(self.env["input"]["fenrir_key"])
        interesting = set(self.fenrir_keys)
        interesting.update(self.env["input"]["script_key"])
        interesting.update(self.modifier_key_names)
        for shortcut in self.env.get("rawBindings", {}).values():
            for key_name in shortcut[1]:
                if key_name == "KEY_FENRIR":
                    interesting.update(self.fenrir_keys)
                elif key_name == "KEY_SCRIPT":
                    interesting.update(self.env["input"]["script_key"])
                else:
                    interesting.add(input_manager.convert_event_name(key_name))
                    interesting.add(key_name)
        self.interesting_keys = interesting

    def refresh_grabs(self, force=False):
        signature = (
            tuple(sorted(self.env.get("rawBindings", {}).keys())),
            tuple(sorted(self.env["input"]["fenrir_key"])),
            tuple(sorted(self.env["input"]["script_key"])),
        )
        if not force and signature == self.grab_signature:
            return
        self.ungrab_all_devices()
        self.refresh_interesting_keys()
        passive_grabs = self.build_passive_grabs()
        failed_before = self.failed_grabs
        for key_name, modifier_mask, include_num_lock in passive_grabs:
            self.grab_key_name(key_name, modifier_mask, include_num_lock)
        self.display.flush()
        self.grab_signature = signature
        self.write_debug(
            "x11Driver passive grabs planned "
            + str(len(passive_grabs))
            + ", installed "
            + str(len(self.grabbed))
            + ", failed "
            + str(self.failed_grabs - failed_before),
            debug.DebugLevel.INFO,
        )

    def build_passive_grabs(self):
        grabs = set()
        for fenrir_key in self.env["input"]["fenrir_key"]:
            grabs.add((fenrir_key, 0, True))
        for script_key in self.env["input"]["script_key"]:
            grabs.add((script_key, 0, True))
        grabs.add(("KEY_NUMLOCK", 0, True))
        for shortcut in self.env.get("rawBindings", {}).values():
            keys = shortcut[1]
            expanded_keys = self.expand_special_keys(keys)
            modifier_mask = self.shortcut_modifier_mask(expanded_keys)
            non_modifier_keys = [
                key for key in expanded_keys
                if key not in self.modifier_key_names
            ]
            if not non_modifier_keys:
                continue
            final_key = non_modifier_keys[-1]
            if "KEY_FENRIR" in keys:
                for fenrir_key in self.env["input"]["fenrir_key"]:
                    grabs.add((fenrir_key, modifier_mask, True))
                    fenrir_modifier_mask = self.modifier_masks.get(
                        fenrir_key, 0
                    )
                    if fenrir_modifier_mask:
                        grabs.add(
                            (
                                final_key,
                                modifier_mask | fenrir_modifier_mask,
                                not final_key.startswith("KEY_KP"),
                            )
                        )
            elif "KEY_SCRIPT" in keys:
                for script_key in self.env["input"]["script_key"]:
                    grabs.add((script_key, modifier_mask, True))
                    script_modifier_mask = self.modifier_masks.get(
                        script_key, 0
                    )
                    if script_modifier_mask:
                        grabs.add(
                            (
                                final_key,
                                modifier_mask | script_modifier_mask,
                                not final_key.startswith("KEY_KP"),
                            )
                        )
            else:
                grabs.add(
                    (
                        final_key,
                        modifier_mask,
                        not final_key.startswith("KEY_KP"),
                    )
                )
        return grabs

    def expand_special_keys(self, keys):
        expanded = []
        for key_name in keys:
            if key_name == "KEY_FENRIR":
                expanded.extend(self.env["input"]["fenrir_key"])
            elif key_name == "KEY_SCRIPT":
                expanded.extend(self.env["input"]["script_key"])
            else:
                expanded.append(key_name)
        return expanded

    def shortcut_modifier_mask(self, keys):
        modifier_mask = 0
        for key_name in keys:
            modifier_mask |= self.modifier_masks.get(key_name, 0)
        return modifier_mask

    def grab_key_name(
        self, key_name, modifier_mask=0, include_num_lock=True
    ):
        keysym_names = self.key_name_to_keysym_names(key_name)
        for keysym_name in keysym_names:
            keysym = XK.string_to_keysym(keysym_name)
            if not keysym:
                continue
            keycode = self.display.keysym_to_keycode(keysym)
            if not keycode:
                continue
            for effective_mask in self.optional_modifier_masks(
                modifier_mask, include_num_lock
            ):
                try:
                    self.window.grab_key(
                        keycode,
                        effective_mask,
                        False,
                        X.GrabModeAsync,
                        X.GrabModeAsync,
                    )
                    self.grabbed.add((keycode, effective_mask))
                except BadAccess:
                    self.failed_grabs += 1
                    self.env["runtime"]["DebugManager"].write_debug_out(
                        "x11Driver could not grab "
                        + key_name
                        + " with modifier mask "
                        + str(effective_mask),
                        debug.DebugLevel.WARNING,
                    )

    def optional_modifier_masks(self, modifier_mask, include_num_lock=True):
        optional_masks = [0, X.LockMask]
        if include_num_lock and self.num_lock_mask:
            optional_masks += [self.num_lock_mask, self.num_lock_mask | X.LockMask]
        return {modifier_mask | optional for optional in optional_masks}

    def key_name_to_keysym_names(self, key_name):
        reverse_map = {
            value: key for key, value in self.key_name_overrides.items()
        }
        aliases = {
            "KEY_META": ["Super_L", "Super_R", "Meta_L", "Meta_R"],
            "KEY_LEFTMETA": ["Super_L", "Meta_L"],
            "KEY_RIGHTMETA": ["Super_R", "Meta_R"],
            "KEY_COMPOSE": ["Multi_key"],
            "KEY_PAGEUP": ["Page_Up", "Prior"],
            "KEY_PAGEDOWN": ["Page_Down", "Next"],
            "KEY_MINUS": ["minus", "underscore"],
            "KEY_EQUAL": ["equal", "plus"],
            "KEY_KP0": ["KP_0", "KP_Insert"],
            "KEY_KP1": ["KP_1", "KP_End"],
            "KEY_KP2": ["KP_2", "KP_Down"],
            "KEY_KP3": ["KP_3", "KP_Page_Down"],
            "KEY_KP4": ["KP_4", "KP_Left"],
            "KEY_KP5": ["KP_5", "KP_Begin"],
            "KEY_KP6": ["KP_6", "KP_Right"],
            "KEY_KP7": ["KP_7", "KP_Home"],
            "KEY_KP8": ["KP_8", "KP_Up"],
            "KEY_KP9": ["KP_9", "KP_Page_Up"],
            "KEY_KPDOT": ["KP_Decimal", "KP_Delete"],
        }
        if key_name in aliases:
            return aliases[key_name]
        if key_name in reverse_map:
            return [reverse_map[key_name]]
        if key_name.startswith("KEY_F") and key_name[5:].isdigit():
            return [key_name[4:]]
        if key_name.startswith("KEY_") and len(key_name) == 5:
            return [key_name[4:].lower()]
        if key_name.startswith("KEY_") and key_name[4:].isdigit():
            return [key_name[4:]]
        return []

    def reset_input_state(self):
        try:
            self.env["runtime"]["InputManager"].reset_input_state()
        except Exception:
            self.clear_event_buffer()

    def write_event_buffer(self):
        self.clear_event_buffer()

    def clear_event_buffer(self):
        if not self._initialized:
            return
        del self.env["input"]["event_buffer"][:]

    def write_debug(self, message, level):
        try:
            self.env["runtime"]["DebugManager"].write_debug_out(message, level)
        except Exception:
            pass

    def update_input_devices(self, new_devices=None, init=False):
        return

    def grab_all_devices(self):
        return True

    def ungrab_all_devices(self):
        if not self.display or not self.window:
            return True
        for keycode, modifier_mask in list(self.grabbed):
            try:
                self.window.ungrab_key(keycode, modifier_mask)
            except Exception:
                pass
        self.grabbed.clear()
        try:
            self.display.flush()
        except Exception:
            pass
        return True

    def remove_all_devices(self):
        self.ungrab_all_devices()

    def get_led_state(self, led=0):
        if led == 0:
            return bool(
                self.num_lock_mask
                and self.modifier_state & self.num_lock_mask
            )
        if led == 1:
            return bool(self.modifier_state & X.LockMask)
        return False

    def set_led_state(self, led_dict):
        return False



#!/usr/bin/env python3

import configparser
import os
import sys
from inspect import isfunction
from os import listdir
from os.path import isfile
from os.path import join

from xdg import BaseDirectory

# Get configuration directory
if len(sys.argv) > 1:
    config_path = sys.argv[1]
elif os.geteuid() == 0:
    # Save settings system wide
    config_path = "/etc/fenrir.conf"
else:
    # Use local settings
    config_path = BaseDirectory.xdg_data_home + "/fenrir"
    if not os.path.exists(config_path):
        os.makedirs(config_path)
    config_path = config_path + "/fenrir.conf"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import time


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return ""

    def run(self):
        if not self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "speech", "auto_read_incoming"
        ):
            return

        if "pendingPromptText" not in self.env["commandBuffer"]:
            return

        pending_text = self.env["commandBuffer"]["pendingPromptText"]
        if not pending_text:
            return

        pending_time = self.env["commandBuffer"].get("pendingPromptTime", 0)
        delay = self.env["runtime"]["SettingsManager"].get_setting_as_float(
            "speech", "batch_flush_interval"
        )
        if time.time() - pending_time < delay:
            return

        self.env["runtime"]["OutputManager"].present_text(
            pending_text, interrupt=False, flush=False
        )
        self.env["commandBuffer"]["pendingPromptText"] = ""

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.


from fenrirscreenreader.core.i18n import _


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return _("selects the last clipboard")

    def run(self):
        if self.env["runtime"]["MemoryManager"].is_index_list_empty(
            "clipboardHistory"
        ):
            self.env["runtime"]["OutputManager"].present_text(
                _("clipboard empty"), interrupt=True
            )
            return
        self.env["runtime"]["MemoryManager"].set_last_index("clipboardHistory")
        clipboard = self.env["runtime"][
            "MemoryManager"
        ].get_index_list_element("clipboardHistory")
        self.env["runtime"]["OutputManager"].present_text(
            clipboard, interrupt=True
        )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core.i18n import _
from fenrirscreenreader.utils import mark_utils


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return _("copies last presented text to the clipboard")

    def run(self):
        last_echo = self.env["runtime"]["OutputManager"].get_last_echo()
        if last_echo.rstrip() != "":
            last_echo = last_echo.rstrip()
        self.env["runtime"]["MemoryManager"].add_value_to_first_index(
            "clipboardHistory", last_echo
        )
        self.env["runtime"]["OutputManager"].present_text(
            last_echo, sound_icon="CopyToClipboard", interrupt=True
        )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.


from fenrirscreenreader.core.i18n import _


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return _("Cycle through key echo modes: character, word, off")

    def run(self):
        settings_manager = self.env["runtime"]["SettingsManager"]
        output_manager = self.env["runtime"]["OutputManager"]

        # Get current settings
        char_echo_mode = settings_manager.get_setting("keyboard", "char_echo_mode")
        word_echo = settings_manager.get_setting_as_bool("keyboard", "word_echo")

        # Determine current state and cycle to next
        # States: character (char=1, word=False) -> word (char=0, word=True) -> off (char=0, word=False)
        if char_echo_mode == "1" and not word_echo:
            # Currently character echo, switch to word echo
            settings_manager.set_setting("keyboard", "char_echo_mode", "0")
            settings_manager.set_setting("keyboard", "word_echo", "True")
            output_manager.present_text(
                _("Echo by word"), interrupt=True
            )
        elif word_echo:
            # Currently word echo, switch to off
            settings_manager.set_setting("keyboard", "char_echo_mode", "0")
            settings_manager.set_setting("keyboard", "word_echo", "False")
            output_manager.present_text(
                _("Echo off"), interrupt=True
            )
        else:
            # Currently off (or caps mode), switch to character echo
            settings_manager.set_setting("keyboard", "char_echo_mode", "1")
            settings_manager.set_setting("keyboard", "word_echo", "False")
            output_manager.present_text(
                _("Echo by character"), interrupt=True
            )

    def set_callback(self, callback):
        pass
